import eventlet
eventlet.monkey_patch(socket=True, select=True, time=True)

import sys
import time

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import service as common_service
from oslo_utils import excutils

from neutron._i18n import _
from neutron._i18n import _LE
from neutron.common import config
from neutron.common import rpc as n_rpc
from neutron.db import api as session
from neutron import service

service_opts = [
    cfg.StrOpt('topic',
               help=_('Topic for messaging to pub/sub to')),
    cfg.StrOpt('transport_url',
               help=_('Connection string for transport service')),
    cfg.IntOpt('periodic_interval',
               default=40,
               help=_('Seconds between running periodic tasks')),
    cfg.IntOpt('rpc_workers',
               default=1,
               help=_('Number of RPC worker processes for service')),
    cfg.IntOpt('periodic_fuzzy_delay',
               default=5,
               help=_('Range of seconds to randomly delay when starting the '
                      'periodic task scheduler to reduce stampeding. '
                      '(Disable by setting to 0)')),
]
CONF = cfg.CONF
CONF.register_opts(service_opts, "QUARK_ASYNC")
LOG = logging.getLogger(__name__)


class QuarkRpcTestCallback(object):
    target = messaging.Target(version='1.0', namespace=None)

    def stuff(self, context, **kwargs):
        return {"status": "okay"}


class QuarkAsyncPlugin(object):
    def __init__(self):
        pass

    def _setup_rpc(self):
        self.endpoints = [QuarkRpcTestCallback()]

    def start_rpc_listeners(self):
        """Configure all listeners here"""
        self._setup_rpc()
        self.conn = n_rpc.create_connection()
        self.conn.create_consumer(CONF.QUARK_ASYNC.topic, self.endpoints,
                                  fanout=False)
        return self.conn.consume_in_threads()


def serve_rpc():

    if cfg.CONF.QUARK_ASYNC.rpc_workers < 1:
        cfg.CONF.set_override('rpc_workers', 1, "QUARK_ASYNC")

    try:
        plugins = [QuarkAsyncPlugin()]
        rpc = service.RpcWorker(plugins)
        session.dispose()  # probaby not needed, but maybe
        launcher = common_service.ProcessLauncher(CONF, wait_interval=1.0)
        launcher.launch_service(rpc, workers=CONF.QUARK_ASYNC.rpc_workers)

        return launcher
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.exception(_LE('Unrecoverable error: please check log for '
                              'details.'))


def start_api_and_rpc_workers():
    pool = eventlet.GreenPool()

    quark_rpc = serve_rpc()
    pool.spawn(quark_rpc.wait)

    pool.waitall()


def boot_server(server_func):
    # the configuration will be read into the cfg.CONF global data structure
    config.init(sys.argv[1:])
    config.setup_logging()
    config.set_config_defaults()
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))
    try:
        server_func()
    except KeyboardInterrupt:
        pass
    except RuntimeError as e:
        sys.exit(_("ERROR: %s") % e)


def main():
    boot_server(start_api_and_rpc_workers)


class QuarkRpcTestApi(object):
    """This class is used for testing QuarkRpcTestCallback."""
    def __init__(self):
        target = messaging.Target(topic=CONF.QUARK_ASYNC.topic)
        self.client = n_rpc.get_client(target)

    def stuff(self, context):
        cctxt = self.client.prepare(version='1.0')
        return cctxt.call(context, 'stuff')


class QuarkAsyncTestContext(object):
    """This class is used for testing QuarkRpcTestCallback."""
    def __init__(self):
        self.time = time.ctime()

    def to_dict(self):
        return {"application": "rpc-client", "time": time.ctime()}


def test_main():
    config.init(sys.argv[1:])
    config.setup_logging()
    config.set_config_defaults()
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))
    context = QuarkAsyncTestContext()  # typically context is neutron context
    client = QuarkRpcTestApi()
    LOG.info(client.stuff(context))
    time.sleep(0)  # necessary for preventing Timeout exceptions


if __name__ == "__main__":
    main()



"""Added locks tables

Revision ID: 341d2e702dc4
Revises: e249ebc4f51
Create Date: 2015-09-03 09:24:29.941684

"""

# revision identifiers, used by Alembic.
revision = '341d2e702dc4'
down_revision = 'e249ebc4f51'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('quark_locks',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('type', sa.Enum('ip_address'), nullable=False),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')
    op.create_table('quark_lock_holders',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('lock_id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['lock_id'], ['quark_locks.id'], ),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')
    op.add_column(u'quark_ip_addresses', sa.Column('lock_id', sa.Integer(),
                                                   nullable=True))


def downgrade():
    op.drop_column(u'quark_ip_addresses', 'lock_id')
    op.drop_table('quark_lock_holders')
    op.drop_table('quark_locks')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

from sqlalchemy.dialects import mysql, sqlite

from quark.db import custom_types
from quark.tests import test_base


class TestDBCustomTypesINET(test_base.TestBase):
    """Adding for coverage of the custom types."""

    def setUp(self):
        super(TestDBCustomTypesINET, self).setUp()
        self.inet = custom_types.INET()

    def test_inet_load_dialect_impl(self):
        dialect = self.inet.load_dialect_impl(mysql.dialect())
        self.assertEqual(type(dialect), type(custom_types.INET.impl()))

    def test_inet_load_dialect_impl_sqlite(self):
        dialect = self.inet.load_dialect_impl(sqlite.dialect())
        self.assertEqual(type(dialect), sqlite.CHAR)

    def test_process_bind_param(self):
        bind = self.inet.process_bind_param(None, None)
        self.assertIsNone(bind)

    def test_process_bind_param_with_value(self):
        bind = self.inet.process_bind_param("foo", sqlite.dialect())
        self.assertEqual(bind, "foo")

    def test_process_bind_param_with_value_not_sqlite(self):
        bind = self.inet.process_bind_param("foo", mysql.dialect())
        self.assertEqual(bind, "foo")

    def test_process_result_value(self):
        bind = self.inet.process_result_value(None, mysql.dialect())
        self.assertIsNone(bind)

    def test_process_result_value_with_value(self):
        bind = self.inet.process_result_value(1.0, sqlite.dialect())
        self.assertEqual(bind, 1.0)

    def test_process_result_value_with_value_not_sqlite(self):
        bind = self.inet.process_result_value(1.0, mysql.dialect())
        self.assertEqual(bind, 1.0)


class TestDBCustomTypesMACAddress(test_base.TestBase):
    """Adding for coverage of the mac address custom types."""

    def setUp(self):
        super(TestDBCustomTypesMACAddress, self).setUp()
        self.mac = custom_types.MACAddress()

    def test_mac_load_dialect_impl(self):
        dialect = self.mac.load_dialect_impl(sqlite.dialect())
        self.assertEqual(type(dialect), sqlite.CHAR)

    def test_mac_load_dialect_impl_not_sqlite(self):
        dialect = self.mac.load_dialect_impl(mysql.dialect())
        self.assertEqual(type(dialect), type(custom_types.MACAddress.impl()))



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

import json

import mock
import netaddr
from oslo_config import cfg
import redis

from quark.agent.xapi import VIF
from quark.cache import security_groups_client as sg_client
from quark.db import models
from quark.environment import Capabilities
from quark import exceptions as q_exc
from quark.tests import test_base

CONF = cfg.CONF


class TestRedisSecurityGroupsClient(test_base.TestBase):

    def setUp(self):
        super(TestRedisSecurityGroupsClient, self).setUp()
        # Forces the connection pool to be recreated on every test
        sg_client.SecurityGroupsClient.connection_pool = None
        temp_envcaps = [Capabilities.SECURITY_GROUPS, Capabilities.EGRESS]
        CONF.set_override('environment_capabilities', temp_envcaps, 'QUARK')

    def tearDown(self):
        CONF.clear_override('environment_capabilities', 'QUARK')

    @mock.patch("uuid.uuid4")
    @mock.patch("quark.cache.redis_base.TwiceRedis")
    def test_apply_rules(self, strict_redis, uuid4):
        client = sg_client.SecurityGroupsClient()
        device_id = "device"
        uuid4.return_value = "uuid"

        mac_address = netaddr.EUI("AA:BB:CC:DD:EE:FF")
        client.apply_rules(device_id, mac_address.value, [])
        self.assertTrue(client._client.master.hset.called)

        redis_key = client.vif_key(device_id, mac_address.value)

        rule_dict = {"rules": []}

        client._client.master.hset.assert_any_call(
            redis_key, sg_client.SECURITY_GROUP_HASH_ATTR,
            json.dumps(rule_dict))

        client._client.master.hset.assert_any_call(
            redis_key, sg_client.SECURITY_GROUP_ACK, False)

    @mock.patch("uuid.uuid4")
    @mock.patch("quark.cache.redis_base.TwiceRedis")
    def test_delete_vif(self, strict_redis, uuid4):
        client = sg_client.SecurityGroupsClient()
        device_id = "device"
        uuid4.return_value = "uuid"
        mac_address = netaddr.EUI("AA:BB:CC:DD:EE:FF")

        redis_key = client.vif_key(device_id, mac_address.value)
        client.delete_vif(device_id, mac_address)
        client._client.master.delete.assert_called_with(redis_key)

    def test_apply_rules_set_fails_gracefully(self):
        port_id = 1
        mac_address = netaddr.EUI("AA:BB:CC:DD:EE:FF")
        conn_err = redis.ConnectionError
        with mock.patch("quark.cache.security_groups_client."
                        "redis_base.ClientBase") as redis_mock:
            mocked_redis_cli = mock.MagicMock()
            redis_mock.return_value = mocked_redis_cli

            client = sg_client.SecurityGroupsClient()
            mocked_redis_cli.master.hset.side_effect = conn_err
            with self.assertRaises(q_exc.RedisConnectionFailure):
                client.apply_rules(port_id, mac_address.value, [])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_group_no_rules(self, strict_redis):
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        payload = client.serialize_groups([group])
        self.assertEqual([], payload)

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_group_with_rules(self, strict_redis):
        rule_dict = {"ethertype": 0x800, "protocol": 6, "port_range_min": 80,
                     "port_range_max": 443, "direction": "ingress"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x800, rule["ethertype"])
        self.assertEqual(6, rule["protocol"])
        self.assertEqual(80, rule["port start"])
        self.assertEqual(443, rule["port end"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("ingress", rule["direction"])
        self.assertEqual("", rule["source network"])
        self.assertEqual("", rule["destination network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_group_with_rules_and_remote_network(self, strict_redis):
        rule_dict = {"ethertype": 0x800, "protocol": 1, "direction": "ingress",
                     "remote_ip_prefix": "192.168.0.0/24"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x800, rule["ethertype"])
        self.assertEqual(1, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("ingress", rule["direction"])
        self.assertEqual("::ffff:192.168.0.0/120", rule["source network"])
        self.assertEqual("", rule["destination network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_group_egress_rules(self, strict_redis):
        rule_dict = {"ethertype": 0x800, "protocol": 1,
                     "direction": "egress",
                     "remote_ip_prefix": "192.168.0.0/24"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x800, rule["ethertype"])
        self.assertEqual(1, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("egress", rule["direction"])
        self.assertEqual("::ffff:192.168.0.0/120", rule["destination network"])
        self.assertEqual("", rule["source network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_filters_source_v4_net(self, strict_redis):
        rule_dict = {"ethertype": 0x800, "protocol": 1, "direction": "ingress",
                     "remote_ip_prefix": "192.168.0.0/0"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x800, rule["ethertype"])
        self.assertEqual(1, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("ingress", rule["direction"])
        self.assertEqual("", rule["source network"])
        self.assertEqual("", rule["destination network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_filters_source_v6_net(self, strict_redis):
        rule_dict = {"ethertype": 0x86DD, "protocol": 58,
                     "direction": "ingress",
                     "remote_ip_prefix": "feed::/0"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x86DD, rule["ethertype"])
        self.assertEqual(58, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("ingress", rule["direction"])
        self.assertEqual("", rule["source network"])
        self.assertEqual("", rule["destination network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_filters_dest_v4_net(self, strict_redis):
        rule_dict = {"ethertype": 0x800, "protocol": 1, "direction": "egress",
                     "remote_ip_prefix": "192.168.0.0/0"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x800, rule["ethertype"])
        self.assertEqual(1, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("egress", rule["direction"])
        self.assertEqual("", rule["source network"])
        self.assertEqual("", rule["destination network"])

    @mock.patch(
        "quark.cache.security_groups_client.redis_base.TwiceRedis")
    def test_serialize_filters_dest_v6_net_(self, strict_redis):
        rule_dict = {"ethertype": 0x86DD, "protocol": 58,
                     "direction": "egress",
                     "remote_ip_prefix": "feed::/0"}
        client = sg_client.SecurityGroupsClient()
        group = models.SecurityGroup()
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group.rules.append(rule)

        payload = client.serialize_groups([group])
        rule = payload[0]
        self.assertEqual(0x86DD, rule["ethertype"])
        self.assertEqual(58, rule["protocol"])
        self.assertEqual(None, rule["icmp type"])
        self.assertEqual(None, rule["icmp code"])
        self.assertEqual("allow", rule["action"])
        self.assertEqual("egress", rule["direction"])
        self.assertEqual("", rule["source network"])
        self.assertEqual("", rule["destination network"])


class TestRedisForAgent(test_base.TestBase):
    def setUp(self):
        super(TestRedisForAgent, self).setUp()

        patch = mock.patch("quark.cache.security_groups_client.redis_base."
                           "TwiceRedis")
        self.MockSentinel = patch.start()
        self.addCleanup(patch.stop)

    @mock.patch(
        "quark.cache.security_groups_client.SecurityGroupsClient.get_fields")
    def test_get_security_group_states_empty(self, mock_get_fields):
        rc = sg_client.SecurityGroupsClient()
        mock_get_fields.return_value = []
        group_states = rc.get_security_group_states([])
        mock_get_fields.assert_called_once_with([],
                                                sg_client.SECURITY_GROUP_ACK)
        self.assertEqual(group_states, {})

    @mock.patch(
        "quark.cache.security_groups_client.SecurityGroupsClient.get_fields")
    def test_get_security_group_states_nonempty(self, mock_get_fields):
        rc = sg_client.SecurityGroupsClient()

        mock_get_fields.return_value = [
            None,
            '{}',
            '{"%s": False}' % sg_client.SECURITY_GROUP_ACK,
            '{"%s": True}' % sg_client.SECURITY_GROUP_ACK,
            '{"%s": "1-2-3"}' % sg_client.SECURITY_GROUP_ACK]

        recs = [{"MAC": 2}, {"MAC": 4}, {"MAC": 6}, {"MAC": 8}, {"MAC": 0}]
        new_interfaces = ([VIF(1, recs[0], 9), VIF(3, recs[1], 0),
                           VIF(5, recs[2], 1), VIF(7, recs[3], 2),
                           VIF(9, recs[4], 3)])

        group_states = rc.get_security_group_states(new_interfaces)

        mock_get_fields.assert_called_once_with(
            ["1.000000000002", "3.000000000004", "5.000000000006",
             "7.000000000008", "9.000000000000"],
            sg_client.SECURITY_GROUP_ACK)

        self.assertEqual(group_states, {new_interfaces[2]: False,
                                        new_interfaces[3]: True})



# Copyright (c) 2016 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron import context
from neutron_lib import exceptions as n_exc

from quark import plugin
from quark.tests.functional.base import BaseFunctionalTest


class QuarkPluginNegativeTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkPluginNegativeTest, self).setUp()
        self.plugin = plugin.Plugin()
        self.blank_context = context.Context(None, None, is_admin=True)

        self.with_tenant_id = {"thing": {"tenant_id": "stuff"}}
        self.without_tenant_id = {"thing": {"attr": "stuff"}}


class QuarkPluginTenantlessNegativeTests(QuarkPluginNegativeTest):

    def test_tenant_check_no_raise(self):
        ret = self.plugin._fix_missing_tenant_id(
            self.blank_context, self.with_tenant_id, "thing")
        self.assertEqual(None, ret)

    def test_tenant_check_raises_if_no_tenant(self):
        with self.assertRaises(n_exc.BadRequest):
            self.plugin._fix_missing_tenant_id(
                self.blank_context, self.without_tenant_id, "thing")

    def test_tenant_check_no_raise_if_tenant_in_context(self):
        self.plugin._fix_missing_tenant_id(
            self.context, self.without_tenant_id, "thing")

    def test_tenant_check_raises_missing_body(self):
        with self.assertRaises(n_exc.BadRequest):
            self.plugin._fix_missing_tenant_id(
                self.blank_context, {}, "thing")

        with self.assertRaises(n_exc.BadRequest):
            self.plugin._fix_missing_tenant_id(
                self.blank_context, None, "thing")



from neutron import context
from neutron.db import api as neutron_db_api
from oslo_config import cfg
from sqlalchemy.orm import configure_mappers

from quark.db import models
from quark import quota_driver
from quark.tests import test_base


class BaseFunctionalTest(test_base.TestBase):
    def setUp(self):
        super(BaseFunctionalTest, self).setUp()
        self.context = context.Context('fake', 'fake', is_admin=False)
        cfg.CONF.set_override('connection', 'sqlite://', 'database')
        configure_mappers()
        self.engine = neutron_db_api.get_engine()
        models.BASEV2.metadata.create_all(self.engine)
        quota_driver.Quota.metadata.create_all(self.engine)

    def tearDown(self):
        self.context = None
        models.BASEV2.metadata.drop_all(self.engine)
        quota_driver.Quota.metadata.drop_all(self.engine)



from quark.db import models
from quark.drivers import optimized_nvp_driver  # noqa
from quark import quota_driver


# add your model's MetaData object here
# for 'autogenerate' support
target_metadata = models.BASEV2.metadata
# FIXME: https://bitbucket.org/zzzeek/alembic/issue/38
table_names = set([tbl.name for tbl in target_metadata.sorted_tables])
for t in quota_driver.Quota.metadata.tables.values():
    if t.name == "quotas" and t.name not in table_names:
        t.tometadata(target_metadata)



import mock
import netaddr

import contextlib

from quark.db import api as db_api
from quark.db import ip_types
import quark.ipam
import quark.plugin
import quark.plugin_modules.ip_addresses as ip_api
import quark.plugin_modules.mac_address_ranges as macrng_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.ports as port_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class QuarkSharedIPs(MySqlBaseFunctionalTest):
    def __init__(self, *args, **kwargs):
        super(QuarkSharedIPs, self).__init__(*args, **kwargs)
        self.cidr = "192.168.2.0/24"
        self.ip_network = netaddr.IPNetwork(self.cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        self.network = {"network": network}
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=self.cidr, first_ip=self.ip_network.first,
                      last_ip=self.ip_network.last, ip_policy=None,
                      tenant_id="fake")
        self.subnet = {"subnet": subnet}
        port1 = {'port': dict(device_id='a')}
        port2 = {'port': dict(device_id='b')}
        port3 = {'port': dict(device_id='c')}
        port4 = {'port': dict(device_id='d')}
        self.ports_info2 = [port1, port2]
        self.ports_info4 = [port1, port2, port3, port4]

    def setUp(self):
        super(QuarkSharedIPs, self).setUp()

    def tearDown(self):
        super(QuarkSharedIPs, self).tearDown()

    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info, ports_info):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            self.context.is_admin = True
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            subnet_info['subnet']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet_info)
            ports = []
            for port_info in ports_info:
                port_info['port']['network_id'] = net['id']
                ports.append(port_api.create_port(self.context, port_info))
            yield net, sub, ports

    def test_create_shared_ips_with_port_ids(self):

        def _make_body(ip):
            fix_ip = dict(ip_address=ip, subnet_id=sub['id'])
            port_info = {"port": dict(fixed_ips=[fix_ip])}
            return port_info

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):
            for p in ports:
                port_db = db_api.port_find(self.context, id=p['id'],
                                           scope=db_api.ONE)
                assocs = db_api.ip_port_association_find(self.context,
                                                         scope=db_api.ALL,
                                                         port_id=p['id'])
                self.assertEqual(1, len(p.get('fixed_ips')))
                self.assertEqual(1, len(port_db.ip_addresses))
                ip_db = port_db.ip_addresses[0]
                self.assertEqual('none', ip_db.get_service_for_port(port_db))
                self.assertEqual(1, len(assocs))

            port_ids = [ports[0]['id'], ports[1]['id']]
            shared_ip = {'ip_address': dict(port_ids=port_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

    def test_shared_ip_in_fixed_ip_list(self):

        def _make_body(service):
            body = dict(service=service)
            port_info = {"port": dict(body)}
            return port_info

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):
            for p in ports:
                port_db = db_api.port_find(self.context, id=p['id'],
                                           scope=db_api.ONE)
                assocs = db_api.ip_port_association_find(self.context,
                                                         scope=db_api.ALL,
                                                         port_id=p['id'])
                self.assertEqual(1, len(p.get('fixed_ips')))
                self.assertEqual(1, len(port_db.ip_addresses))
                ip_db = port_db.ip_addresses[0]
                self.assertEqual('none', ip_db.get_service_for_port(port_db))
                self.assertEqual(1, len(assocs))

            device_ids = [ports[0]['device_id'], ports[1]['device_id']]
            shared_ip = {'ip_address': dict(device_ids=device_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

            port = port_api.get_port(self.context, ports[0]['id'])
            self.assertEqual(2, len(port['fixed_ips']))

            port_ip_update = ip_api.update_port_for_ip_address
            updated_port = port_ip_update(self.context, ip['id'],
                                          ports[0]['id'], _make_body('derp'))
            self.assertEqual('derp', updated_port.get('service'))

            port = ip_api.get_port_for_ip_address(self.context, ip['id'],
                                                  ports[0]['id'])
            self.assertEqual('derp', port.get('service'))
            port = ip_api.get_port_for_ip_address(self.context, ip['id'],
                                                  ports[1]['id'])
            self.assertEqual('none', port.get('service'))

    def test_create_shared_ips_with_device_ids(self):

        with self._stubs(self.network, self.subnet, self.ports_info2) as (
                net, sub, ports):
            for p in ports:
                port_db = db_api.port_find(self.context, id=p['id'],
                                           scope=db_api.ONE)
                assocs = db_api.ip_port_association_find(self.context,
                                                         scope=db_api.ALL,
                                                         port_id=p['id'])
                self.assertEqual(1, len(p.get('fixed_ips')))
                self.assertEqual(1, len(port_db.ip_addresses))
                ip_db = port_db.ip_addresses[0]
                self.assertEqual('none', ip_db.get_service_for_port(port_db))
                self.assertEqual(1, len(assocs))

            device_ids = [ports[0]['device_id'], ports[1]['device_id']]
            shared_ip = {'ip_address': dict(device_ids=device_ids,
                                            network_id=net['id'],
                                            version=4)}
            ip = ip_api.create_ip_address(self.context, shared_ip)
            self.assertEqual(ip_types.SHARED, ip['type'])

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip['id'])
            self.assertEqual(2, len(ports_ip))

    def test_filter_ip_by_device_and_service(self):

        def _make_body(service):
            body = dict(service=service)
            port_info = {"port": dict(body)}
            return port_info

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):
            port_ids1 = [ports[0]['id'], ports[1]['id']]
            port_ids2 = [ports[2]['id'], ports[3]['id']]

            filters = dict(device_id='a')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(1, len(ips))

            shared_ip1 = {'ip_address': dict(port_ids=port_ids1,
                                             network_id=net['id'],
                                             version=4)}
            ip1 = ip_api.create_ip_address(self.context, shared_ip1)
            self.assertEqual(2, len(ip1['port_ids']))

            shared_ip2 = {'ip_address': dict(port_ids=port_ids2,
                                             network_id=net['id'],
                                             version=4)}
            ip2 = ip_api.create_ip_address(self.context, shared_ip2)
            self.assertEqual(2, len(ip2['port_ids']))

            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip1['id'])
            self.assertEqual(2, len(ports_ip))
            ports_ip = ip_api.get_ports_for_ip_address(self.context, ip2['id'])
            self.assertEqual(2, len(ports_ip))

            filters = dict(device_id='a')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(2, len(ips))

            filters = dict(device_id='x')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(0, len(ips))

            filters = dict(device_id='a', service='derp')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(0, len(ips))

            filters = dict(service='derp')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(0, len(ips))

            filters = dict(device_id='a', service='none')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(2, len(ips))

            port_ip_update = ip_api.update_port_for_ip_address
            updated_port = port_ip_update(self.context, ip1['id'],
                                          ports[0]['id'], _make_body('derp'))
            self.assertEqual('derp', updated_port.get('service'))

            filters = dict(device_id='a', service='derp')
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(1, len(ips))

            filters = dict(device_id='a', service='derp',
                           type=ip_types.FIXED)
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(0, len(ips))

            filters = dict(device_id='a', service='derp',
                           type=ip_types.SHARED)
            ips = ip_api.get_ip_addresses(self.context, **filters)
            self.assertEqual(1, len(ips))

    def test_get_ports_filter_with_ip_and_device(self):

        with self._stubs(self.network, self.subnet, self.ports_info4) as (
                net, sub, ports):

            network = dict(name="xx", tenant_id="fake", network_plugin="BASE")
            xx_network = {"network": network}
            xx_net = network_api.create_network(self.context, xx_network)
            subnet = dict(id=2, ip_version=4, next_auto_assign_ip=2,
                          cidr=self.cidr, first_ip=self.ip_network.first,
                          last_ip=self.ip_network.last, ip_policy=None,
                          tenant_id="fake")
            xx_subnet = {"subnet": subnet}
            xx_subnet['subnet']['network_id'] = xx_net['id']
            subnet_api.create_subnet(self.context, xx_subnet)

            port_info = {'port': dict(device_id='a')}
            port_info['port']['network_id'] = xx_net['id']
            port_api.create_port(self.context, port_info)

            port_ids1 = [ports[0]['id'], ports[1]['id']]

            shared_ip1 = {'ip_address': dict(port_ids=port_ids1,
                                             network_id=net['id'],
                                             version=4)}
            ip1 = ip_api.create_ip_address(self.context, shared_ip1)

            filters = dict(device_id='a')
            ports = ip_api.get_ports_for_ip_address(self.context, ip1['id'],
                                                    filters=filters)
            self.assertEqual(1, len(ports))

            filters = dict(device_id='a')
            ports = port_api.get_ports(self.context, filters=filters)
            self.assertEqual(2, len(ports))


class QuarkIPAddressFind(MySqlBaseFunctionalTest):
    def setUp(self):
        super(QuarkIPAddressFind, self).setUp()
        self.address = netaddr.IPAddress("0.0.0.1")
        db_api.ip_address_create(self.context, address=self.address)
        self.context.session.flush()

    def test_ip_address_find_address_integer(self):
        result = db_api.ip_address_find(
            self.context, address=int(self.address), scope=db_api.ONE)
        self.assertEqual(result.address_readable, str(self.address))

    def test_ip_address_find_address_string(self):
        result = db_api.ip_address_find(
            self.context, address=str(self.address), scope=db_api.ONE)
        self.assertEqual(result.address, int(self.address.ipv6()))



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
import netaddr
from neutron.common import rpc
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import api as db_api
import quark.ipam
import quark.plugin
from quark.tests.functional.base import BaseFunctionalTest


class QuarkNetworkFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkNetworkFunctionalTest, self).setUp()

        patcher = mock.patch("neutron.common.rpc.oslo_messaging")
        patcher.start()
        self.addCleanup(patcher.stop)
        rpc.init(mock.MagicMock())


class QuarkGetNetwork(QuarkNetworkFunctionalTest):

    def test_show_ipam_strategy(self):
        plugin = quark.plugin.Plugin()
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = dict(network=network)
        original = cfg.CONF.QUARK.show_ipam_strategy

        cfg.CONF.set_override('show_ipam_strategy', True, "QUARK")
        net = plugin.create_network(self.context, network)
        self.assertTrue('ipam_strategy' in net)
        net = plugin.get_network(self.context, net['id'])
        self.assertTrue('ipam_strategy' in net)

        cfg.CONF.set_override('show_ipam_strategy', False, "QUARK")
        net = plugin.create_network(self.context, network)
        self.assertFalse('ipam_strategy' in net)
        net = plugin.get_network(self.context, net['id'])
        self.assertFalse('ipam_strategy' in net)

        cfg.CONF.set_override('show_ipam_strategy', original, "QUARK")


class QuarkDeleteNetworKDeallocatedIPs(QuarkNetworkFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet, dealloc=True):
        self.plugin = quark.plugin.Plugin()
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            next_auto = subnet.pop("next_auto_assign_ip", 0)
            sub_mod = db_api.subnet_create(self.context, **subnet)
            db_api.subnet_update(self.context,
                                 sub_mod,
                                 next_auto_assign_ip=next_auto)

        ip_addr = []
        self.ipam.allocate_ip_address(self.context, ip_addr,
                                      net_mod["id"], 0, 0)
        if dealloc:
            self.ipam.deallocate_ip_address(self.context, ip_addr[0])
        yield net_mod

    def test_delete_network_with_allocated_ips_fails(self):
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2

        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=next_ip,
                      cidr="0.0.0.0/24",
                      ip_policy=None, tenant_id="fake")
        with self._stubs(network, subnet, dealloc=False) as net_mod:
            with self.assertRaises(n_exc.SubnetInUse):
                self.plugin.delete_network(self.context, net_mod["id"])

    def test_delete_network_with_deallocated_ips(self):
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        ipnet = netaddr.IPNetwork("0.0.0.0/24")
        next_ip = ipnet.ipv6().first + 2
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=next_ip,
                      cidr="0.0.0.0/24",
                      ip_policy=None, tenant_id="fake")
        with self._stubs(network, subnet) as net_mod:
            try:
                self.plugin.delete_network(self.context, net_mod["id"])
            except Exception:
                self.fail("delete network raised")



"""add_port_network_plugin

Revision ID: 374c1bdb4480
Revises: 4da4444d7706
Create Date: 2015-10-20 12:08:24.780056

"""

# revision identifiers, used by Alembic.
revision = '374c1bdb4480'
down_revision = '4da4444d7706'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('quark_ports', sa.Column('network_plugin',
                  sa.String(length=36), nullable=True))


def downgrade():
    op.drop_column('quark_ports', 'network_plugin')



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.common import exceptions as n_exc_ext
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class AllocationPools(object):
    def __init__(self, subnet_cidr, pools=None, policies=None):
        self._exclude_cidrs = None
        self._subnet_cidr = netaddr.IPNetwork(subnet_cidr)
        self._alloc_pools = pools
        version = self._subnet_cidr.version
        self._subnet_first_ip = netaddr.IPAddress(
            self._subnet_cidr.first, version=version)
        self._subnet_last_ip = netaddr.IPAddress(
            self._subnet_cidr.last, version=version)

        # If passed an empty list, the entire subnet is unallocatable.
        # If None is passed, the entire subnet is allocatable
        if self._alloc_pools is None:
            self._alloc_pools = [
                {"start": self._subnet_first_ip,
                 "end": self._subnet_last_ip}]
        self._policies = policies or []

    def __len__(self):
        return len(self._alloc_pools)

    # Note(asadoughi): Copied from neutron/db/db_base_plugin_v2.py
    def _validate_allocation_pools(self):
        """Validate IP allocation pools.

        Verify start and end address for each allocation pool are valid,
        ie: constituted by valid and appropriately ordered IP addresses.
        Also, verify pools do not overlap among themselves.
        Finally, verify that each range fall within the subnet's CIDR.
        """
        ip_pools = self._alloc_pools
        subnet_cidr = self._subnet_cidr

        LOG.debug(_("Performing IP validity checks on allocation pools"))
        ip_sets = []
        for ip_pool in ip_pools:
            try:
                start_ip = netaddr.IPAddress(ip_pool['start'])
                end_ip = netaddr.IPAddress(ip_pool['end'])
            except netaddr.AddrFormatError:
                LOG.info(_("Found invalid IP address in pool: "
                           "%(start)s - %(end)s:"),
                         {'start': ip_pool['start'],
                          'end': ip_pool['end']})
                raise n_exc_ext.InvalidAllocationPool(pool=ip_pool)
            if (start_ip.version != self._subnet_cidr.version or
                    end_ip.version != self._subnet_cidr.version):
                LOG.info(_("Specified IP addresses do not match "
                           "the subnet IP version"))
                raise n_exc_ext.InvalidAllocationPool(pool=ip_pool)
            if end_ip < start_ip:
                LOG.info(_("Start IP (%(start)s) is greater than end IP "
                           "(%(end)s)"),
                         {'start': ip_pool['start'], 'end': ip_pool['end']})
                raise n_exc_ext.InvalidAllocationPool(pool=ip_pool)
            if (start_ip < self._subnet_first_ip or
                    end_ip > self._subnet_last_ip):
                LOG.info(_("Found pool larger than subnet "
                           "CIDR:%(start)s - %(end)s"),
                         {'start': ip_pool['start'],
                          'end': ip_pool['end']})
                raise n_exc_ext.OutOfBoundsAllocationPool(
                    pool=ip_pool,
                    subnet_cidr=subnet_cidr)
            # Valid allocation pool
            # Create an IPSet for it for easily verifying overlaps
            ip_sets.append(netaddr.IPSet(netaddr.IPRange(
                ip_pool['start'],
                ip_pool['end']).cidrs()))

        LOG.debug(_("Checking for overlaps among allocation pools "
                    "and gateway ip"))
        ip_ranges = ip_pools[:]

        # Use integer cursors as an efficient way for implementing
        # comparison and avoiding comparing the same pair twice
        for l_cursor in xrange(len(ip_sets)):
            for r_cursor in xrange(l_cursor + 1, len(ip_sets)):
                if ip_sets[l_cursor] & ip_sets[r_cursor]:
                    l_range = ip_ranges[l_cursor]
                    r_range = ip_ranges[r_cursor]
                    LOG.info(_("Found overlapping ranges: %(l_range)s and "
                               "%(r_range)s"),
                             {'l_range': l_range, 'r_range': r_range})
                    raise n_exc_ext.OverlappingAllocationPools(
                        pool_1=l_range,
                        pool_2=r_range,
                        subnet_cidr=subnet_cidr)

    def _build_excludes(self):
        self._validate_allocation_pools()
        subnet_net = netaddr.IPNetwork(self._subnet_cidr)
        version = subnet_net.version
        cidrset = netaddr.IPSet(
            netaddr.IPRange(
                netaddr.IPAddress(subnet_net.first, version=version),
                netaddr.IPAddress(subnet_net.last, version=version)).cidrs())

        if isinstance(self._alloc_pools, list):
            for p in self._alloc_pools:
                start = netaddr.IPAddress(p["start"])
                end = netaddr.IPAddress(p["end"])
                cidrset -= netaddr.IPSet(netaddr.IPRange(
                    netaddr.IPAddress(start),
                    netaddr.IPAddress(end)).cidrs())
        elif self._alloc_pools is None:
            # Empty list is completely unallocatable, None is fully
            # allocatable
            cidrset = netaddr.IPSet()

        for p in self._policies:
            cidrset.add(netaddr.IPNetwork(p))

        self._exclude_cidrs = cidrset

    def _refresh_excludes(self):
        if not self._exclude_cidrs:
            self._build_excludes()

    def add_pool(self, pool):
        self._exclude_cidrs = None
        self._alloc_pools.append(pool)

    def add_policy(self, policy):
        self._exclude_cidrs = None
        self._policies.append(policy)

    def validate_gateway_excluded(self, gateway_ip):
        self._refresh_excludes()
        gateway_ip_addr = netaddr.IPAddress(gateway_ip)
        if gateway_ip_addr in self._subnet_cidr:
            if (not self._exclude_cidrs or
                    (self._exclude_cidrs and gateway_ip_addr
                     not in self._exclude_cidrs)):
                raise n_exc_ext.GatewayConflictWithAllocationPools(
                    ip_address=gateway_ip, pool=self._alloc_pools)

    def get_policy_cidrs(self):
        self._refresh_excludes()
        return [str(c) for c in self._exclude_cidrs.iter_cidrs()]



"""Populate enabled and do_not_use

Revision ID: 3ed0c5a067f1
Revises: 5927940a466e
Create Date: 2014-07-07 07:52:27.989294

"""

# revision identifiers, used by Alembic.
revision = '3ed0c5a067f1'
down_revision = '5927940a466e'

from alembic import op
from sqlalchemy.sql import column, table
import sqlalchemy as sa


def upgrade():
    port_ip_associations = table('quark_port_ip_address_associations',
                                 column('enabled', sa.Boolean()))
    mac_addr_ranges = table('quark_mac_address_ranges',
                            column('do_not_use', sa.Boolean()))

    connection = op.get_bind()

    a = port_ip_associations.update().values({'enabled': True})
    connection.execute(a)

    b = mac_addr_ranges.update().values({'do_not_use': False})
    connection.execute(b)


def downgrade():
    pass



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unicorn driver for Quark
"""

import json
import requests

from oslo_config import cfg
from oslo_log import log as logging

from quark.db import ip_types
from quark import exceptions as ex

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

quark_router_opts = [
    cfg.StrOpt('floating_ip_base_url',
               default='http://localhost:8080/v1.0/floating_ips',
               help=_('floating ips base url')),
    cfg.IntOpt('unicorn_api_timeout_seconds', default=2,
               help=_('Number of seconds to wait for a response from the '
                      'server before failing the call'))
]

CONF.register_opts(quark_router_opts, "QUARK")


class UnicornDriver(object):
    def __init__(self):
        pass

    @classmethod
    def get_name(cls):
        return "Unicorn"

    def register_floating_ip(self, floating_ip, port_fixed_ips):
        """Register a floating ip with Unicorn

        :param floating_ip: The quark.db.models.IPAddress to register
        :param port_fixed_ips: A dictionary containing the port and fixed ips
        to associate the floating IP with.  Has the structure of:
        {"<id of port>": {"port": <quark.db.models.Port>,
         "fixed_ip": "<fixed ip address>"}}
        :return: None
        """
        url = CONF.QUARK.floating_ip_base_url
        timeout = CONF.QUARK.unicorn_api_timeout_seconds
        req = self._build_request_body(floating_ip, port_fixed_ips)

        try:
            LOG.info("Calling unicorn to register floating ip: %s %s"
                     % (url, req))
            r = requests.post(url, data=json.dumps(req), timeout=timeout)
        except Exception as e:
            LOG.error("Unhandled Exception caught when trying to register "
                      "floating ip %s with the unicorn API.  Error: %s"
                      % (floating_ip.id, e.message))
            raise ex.RegisterFloatingIpFailure(id=floating_ip.id)

        if r.status_code != 200 and r.status_code != 201:
            msg = "Unexpected status from unicorn API: Status Code %s, " \
                  "Message: %s" % (r.status_code, r.json())
            LOG.error("register_floating_ip: %s" % msg)
            raise ex.RegisterFloatingIpFailure(id=floating_ip.id)

    def update_floating_ip(self, floating_ip, port_fixed_ips):
        """Update an existing floating ip with Unicorn

        :param floating_ip: The quark.db.models.IPAddress to update
        :param port_fixed_ips: A dictionary containing the port and fixed ips
        to associate the floating IP with.  Has the structure of:
        {"<id of port>": {"port": <quark.db.models.Port>,
         "fixed_ip": "<fixed ip address>"}}
        :return: None
        """
        url = "%s/%s" % (CONF.QUARK.floating_ip_base_url,
                         floating_ip["address_readable"])
        timeout = CONF.QUARK.unicorn_api_timeout_seconds
        req = self._build_request_body(floating_ip, port_fixed_ips)

        try:
            LOG.info("Calling unicorn to update floating ip: %s %s"
                     % (url, req))
            r = requests.put(url, data=json.dumps(req), timeout=timeout)
        except Exception as e:
            LOG.error("Unhandled Exception caught when trying to update "
                      "floating ip %s with the unicorn API.  Error: %s"
                      % (floating_ip.id, e.message))
            raise ex.RegisterFloatingIpFailure(id=floating_ip.id)

        if r.status_code != 200 and r.status_code != 201:
            msg = "Unexpected status from unicorn API: Status Code %s, " \
                  "Message: %s" % (r.status_code, r.json())
            LOG.error("register_floating_ip: %s" % msg)
            raise ex.RegisterFloatingIpFailure(id=floating_ip.id)

    def remove_floating_ip(self, floating_ip):
        """Register a floating ip with Unicorn

        :param floating_ip: The quark.db.models.IPAddress to remove
        :return: None
        """
        url = "%s/%s" % (CONF.QUARK.floating_ip_base_url,
                         floating_ip.address_readable)
        timeout = CONF.QUARK.unicorn_api_timeout_seconds

        try:
            LOG.info("Calling unicorn to remove floating ip: %s" % url)
            r = requests.delete(url, timeout=timeout)
        except Exception as e:
            LOG.error("Unhandled Exception caught when trying to un-register "
                      "floating ip %s with the unicorn API.  Error: %s"
                      % (floating_ip.id, e.message))
            raise ex.RemoveFloatingIpFailure(id=floating_ip.id)

        if r.status_code == 404:
            LOG.warn("The floating IP %s does not exist in the unicorn system."
                     % floating_ip.address_readable)
        elif r.status_code != 204:
            msg = "Unexpected status from unicorn API: Status Code %s, " \
                  "Message: %s" % (r.status_code, r.json())
            LOG.error("remove_floating_ip: %s" % msg)
            raise ex.RemoveFloatingIpFailure(id=floating_ip.id)

    @classmethod
    def _build_fixed_ips(cls, port):
        fixed_ips = [{"ip_address": ip.address_readable,
                      "version": ip.version,
                      "subnet_id": ip.subnet.id,
                      "cidr": ip.subnet.cidr,
                      "address_type": ip.address_type}
                     for ip in port.ip_addresses
                     if (ip.get("address_type") == ip_types.FIXED)]
        return fixed_ips

    @classmethod
    def _build_endpoints(cls, port_fixed_ips):
        endpoints = []
        for port_id in port_fixed_ips:
            port = port_fixed_ips[port_id]["port"]
            fixed_ip = port_fixed_ips[port_id]["fixed_ip"]
            endpoint_port = {"uuid": port.id,
                             "name": port.name,
                             "network_uuid": port.network_id,
                             "mac_address": port.mac_address,
                             "device_id": port.device_id,
                             "device_owner": port.device_owner,
                             "fixed_ip": cls._build_fixed_ips(port)}
            endpoints.append({"port": endpoint_port,
                              "private_ip": fixed_ip.address_readable})
        return endpoints

    @classmethod
    def _build_request_body(cls, floating_ip, port_fixed_ips):
        content = {"public_ip": floating_ip["address_readable"],
                   "endpoints": cls._build_endpoints(port_fixed_ips)}
        return {"floating_ip": content}



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.extensions import securitygroup as sg_ext
from neutron import quota
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

from quark.db import api as db_api
from quark.drivers import registry
from quark.environment import Capabilities
from quark import exceptions as q_exc
from quark import ipam
from quark import network_strategy
from quark import plugin_views as v
from quark import tags
from quark import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
PORT_TAG_REGISTRY = tags.PORT_TAG_REGISTRY
STRATEGY = network_strategy.STRATEGY


# HACK(amir): RM9305: do not allow a tenant to associate a network to a port
# that does not belong to them unless it is publicnet or servicenet
# NOTE(blogan): allow advanced services, such as lbaas, the ability
# to associate a network to a port that does not belong to them
def _raise_if_unauthorized(context, net):
    if (not STRATEGY.is_provider_network(net["id"]) and
            net["tenant_id"] != context.tenant_id and
            not context.is_advsvc):
        raise n_exc.NotAuthorized()


def _get_net_driver(network, port=None):
    port_driver = None
    if port and port.get("network_plugin"):
        port_driver = port.get("network_plugin")

    try:
        return registry.DRIVER_REGISTRY.get_driver(
            network["network_plugin"], port_driver=port_driver)
    except Exception as e:
        raise n_exc.BadRequest(resource="ports",
                               msg="invalid network_plugin: %s" % e)


def _get_ipam_driver(network, port=None):
    network_id = network["id"]
    network_strategy = network["ipam_strategy"]

    # Ask the net driver for a IPAM strategy to use
    # with the given network/default strategy.
    net_driver = _get_net_driver(network, port=port)
    strategy = net_driver.select_ipam_strategy(
        network_id, network_strategy)

    # If the driver has no opinion about which strategy to use,
    # we use the one specified by the network.
    if not strategy:
        strategy = network_strategy

    try:
        return ipam.IPAM_REGISTRY.get_strategy(strategy)
    except Exception as e:
        raise n_exc.BadRequest(resource="ports",
                               msg="invalid ipam_strategy: %s" % e)


# NOTE(morgabra) Backend driver operations return a lot of stuff. We use a
# small subset of this data, so we filter out things we don't care about
# so we can avoid any collisions with real port data.
def _filter_backend_port(backend_port):
    # Collect a list of allowed keys in the driver response
    required_keys = ["uuid", "bridge"]
    tag_keys = [tag for tag in PORT_TAG_REGISTRY.tags.keys()]

    allowed_keys = required_keys + tag_keys
    for k in backend_port.keys():
        if k not in allowed_keys:
            del backend_port[k]


def split_and_validate_requested_subnets(context, net_id, segment_id,
                                         fixed_ips):
    subnets = []
    ip_addresses = {}
    for fixed_ip in fixed_ips:
        subnet_id = fixed_ip.get("subnet_id")
        ip_address = fixed_ip.get("ip_address")
        if not subnet_id:
            raise n_exc.BadRequest(resource="fixed_ips",
                                   msg="subnet_id required")
        if ip_address:
            ip_addresses[ip_address] = subnet_id
        else:
            subnets.append(subnet_id)

    subnets = ip_addresses.values() + subnets

    sub_models = db_api.subnet_find(context, id=subnets, scope=db_api.ALL)
    if len(sub_models) == 0:
        raise n_exc.SubnetNotFound(subnet_id=subnets)

    for s in sub_models:
        if s["network_id"] != net_id:
            raise n_exc.InvalidInput(
                error_message="Requested subnet doesn't belong to requested "
                              "network")

        if segment_id and segment_id != s["segment_id"]:
            raise q_exc.AmbiguousNetworkId(net_id=net_id)

    return ip_addresses, subnets


def create_port(context, port):
    """Create a port

    Create a port which is a connection point of a device (e.g., a VM
    NIC) to attach to a L2 Neutron network.
    : param context: neutron api request context
    : param port: dictionary describing the port, with keys
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.  All keys will be populated.
    """
    LOG.info("create_port for tenant %s" % context.tenant_id)
    port_attrs = port["port"]

    admin_only = ["mac_address", "device_owner", "bridge", "admin_state_up",
                  "use_forbidden_mac_range", "network_plugin",
                  "instance_node_id"]
    utils.filter_body(context, port_attrs, admin_only=admin_only)

    port_attrs = port["port"]
    mac_address = utils.pop_param(port_attrs, "mac_address", None)
    use_forbidden_mac_range = utils.pop_param(port_attrs,
                                              "use_forbidden_mac_range", False)
    segment_id = utils.pop_param(port_attrs, "segment_id")
    fixed_ips = utils.pop_param(port_attrs, "fixed_ips")

    if "device_id" not in port_attrs:
        port_attrs['device_id'] = ""
    device_id = port_attrs['device_id']

    # NOTE(morgabra) This should be instance.node from nova, only needed
    # for ironic_driver.
    if "instance_node_id" not in port_attrs:
        port_attrs['instance_node_id'] = ""
    instance_node_id = port_attrs['instance_node_id']

    net_id = port_attrs["network_id"]

    port_id = uuidutils.generate_uuid()

    net = db_api.network_find(context, None, None, None, False, id=net_id,
                              scope=db_api.ONE)

    if not net:
        raise n_exc.NetworkNotFound(net_id=net_id)
    _raise_if_unauthorized(context, net)

    # NOTE (Perkins): If a device_id is given, try to prevent multiple ports
    # from being created for a device already attached to the network
    if device_id:
        existing_ports = db_api.port_find(context,
                                          network_id=net_id,
                                          device_id=device_id,
                                          scope=db_api.ONE)
        if existing_ports:
            raise n_exc.BadRequest(
                resource="port", msg="This device is already connected to the "
                "requested network via another port")

    # Try to fail early on quotas and save ourselves some db overhead
    if fixed_ips:
        quota.QUOTAS.limit_check(context, context.tenant_id,
                                 fixed_ips_per_port=len(fixed_ips))

    if not STRATEGY.is_provider_network(net_id):
        # We don't honor segmented networks when they aren't "shared"
        segment_id = None
        port_count = db_api.port_count_all(context, network_id=[net_id],
                                           tenant_id=[context.tenant_id])
        quota.QUOTAS.limit_check(
            context, context.tenant_id,
            ports_per_network=port_count + 1)
    else:
        if not segment_id:
            raise q_exc.AmbiguousNetworkId(net_id=net_id)

    network_plugin = utils.pop_param(port_attrs, "network_plugin")
    if not network_plugin:
        network_plugin = net["network_plugin"]
    port_attrs["network_plugin"] = network_plugin

    ipam_driver = _get_ipam_driver(net, port=port_attrs)
    net_driver = _get_net_driver(net, port=port_attrs)
    # NOTE(morgabra) It's possible that we select a driver different than
    # the one specified by the network. However, we still might need to use
    # this for some operations, so we also fetch it and pass it along to
    # the backend driver we are actually using.
    base_net_driver = _get_net_driver(net)

    # TODO(anyone): security groups are not currently supported on port create.
    #               Please see JIRA:NCP-801
    security_groups = utils.pop_param(port_attrs, "security_groups")
    if security_groups is not None:
        raise q_exc.SecurityGroupsNotImplemented()

    group_ids, security_groups = _make_security_group_list(context,
                                                           security_groups)
    quota.QUOTAS.limit_check(context, context.tenant_id,
                             security_groups_per_port=len(group_ids))
    addresses = []
    backend_port = None

    with utils.CommandManager().execute() as cmd_mgr:
        @cmd_mgr.do
        def _allocate_ips(fixed_ips, net, port_id, segment_id, mac,
                          **kwargs):
            if fixed_ips:
                if (STRATEGY.is_provider_network(net_id) and
                        not context.is_admin):
                    raise n_exc.NotAuthorized()

                ips, subnets = split_and_validate_requested_subnets(context,
                                                                    net_id,
                                                                    segment_id,
                                                                    fixed_ips)
                kwargs["ip_addresses"] = ips
                kwargs["subnets"] = subnets

            ipam_driver.allocate_ip_address(
                context, addresses, net["id"], port_id,
                CONF.QUARK.ipam_reuse_after, segment_id=segment_id,
                mac_address=mac, **kwargs)

        @cmd_mgr.undo
        def _allocate_ips_undo(addr, **kwargs):
            LOG.info("Rolling back IP addresses...")
            if addresses:
                for address in addresses:
                    try:
                        with context.session.begin():
                            ipam_driver.deallocate_ip_address(context, address,
                                                              **kwargs)
                    except Exception:
                        LOG.exception("Couldn't release IP %s" % address)

        @cmd_mgr.do
        def _allocate_mac(net, port_id, mac_address,
                          use_forbidden_mac_range=False,
                          **kwargs):
            mac = ipam_driver.allocate_mac_address(
                context, net["id"], port_id, CONF.QUARK.ipam_reuse_after,
                mac_address=mac_address,
                use_forbidden_mac_range=use_forbidden_mac_range, **kwargs)
            return mac

        @cmd_mgr.undo
        def _allocate_mac_undo(mac, **kwargs):
            LOG.info("Rolling back MAC address...")
            if mac:
                try:
                    with context.session.begin():
                        ipam_driver.deallocate_mac_address(context,
                                                           mac["address"])
                except Exception:
                    LOG.exception("Couldn't release MAC %s" % mac)

        @cmd_mgr.do
        def _allocate_backend_port(mac, addresses, net, port_id, **kwargs):
            backend_port = net_driver.create_port(
                context, net["id"],
                port_id=port_id,
                security_groups=group_ids,
                device_id=device_id,
                instance_node_id=instance_node_id,
                mac_address=mac,
                addresses=addresses,
                base_net_driver=base_net_driver)
            _filter_backend_port(backend_port)
            return backend_port

        @cmd_mgr.undo
        def _allocate_back_port_undo(backend_port,
                                     **kwargs):
            LOG.info("Rolling back backend port...")
            try:
                backend_port_uuid = None
                if backend_port:
                    backend_port_uuid = backend_port.get("uuid")
                net_driver.delete_port(context, backend_port_uuid)
            except Exception:
                LOG.exception(
                    "Couldn't rollback backend port %s" % backend_port)

        @cmd_mgr.do
        def _allocate_db_port(port_attrs, backend_port, addresses, mac,
                              **kwargs):
            port_attrs["network_id"] = net["id"]
            port_attrs["id"] = port_id
            port_attrs["security_groups"] = security_groups

            LOG.info("Including extra plugin attrs: %s" % backend_port)
            port_attrs.update(backend_port)
            with context.session.begin():
                new_port = db_api.port_create(
                    context, addresses=addresses, mac_address=mac["address"],
                    backend_key=backend_port["uuid"], **port_attrs)

            return new_port

        @cmd_mgr.undo
        def _allocate_db_port_undo(new_port,
                                   **kwargs):
            LOG.info("Rolling back database port...")
            if not new_port:
                return
            try:
                with context.session.begin():
                    db_api.port_delete(context, new_port)
            except Exception:
                LOG.exception(
                    "Couldn't rollback db port %s" % backend_port)

        # addresses, mac, backend_port, new_port
        mac = _allocate_mac(net, port_id, mac_address,
                            use_forbidden_mac_range=use_forbidden_mac_range)
        _allocate_ips(fixed_ips, net, port_id, segment_id, mac)
        backend_port = _allocate_backend_port(mac, addresses, net, port_id)
        new_port = _allocate_db_port(port_attrs, backend_port, addresses, mac)

    return v._make_port_dict(new_port)


def update_port(context, id, port):
    """Update values of a port.

    : param context: neutron api request context
    : param id: UUID representing the port to update.
    : param port: dictionary with keys indicating fields to update.
        valid keys are those that have a value of True for 'allow_put'
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.
    """
    LOG.info("update_port %s for tenant %s" % (id, context.tenant_id))
    port_db = db_api.port_find(context, id=id, scope=db_api.ONE)
    if not port_db:
        raise n_exc.PortNotFound(port_id=id)

    port_dict = port["port"]
    fixed_ips = port_dict.pop("fixed_ips", None)

    admin_only = ["mac_address", "device_owner", "bridge", "admin_state_up",
                  "device_id"]
    always_filter = ["network_id", "backend_key", "network_plugin"]
    utils.filter_body(context, port_dict, admin_only=admin_only,
                      always_filter=always_filter)

    # Pre-check the requested fixed_ips before making too many db trips.
    # Note that this is the only check we need, since this call replaces
    # the entirety of the IP addresses document if fixed_ips are provided.
    if fixed_ips:
        quota.QUOTAS.limit_check(context, context.tenant_id,
                                 fixed_ips_per_port=len(fixed_ips))

    new_security_groups = utils.pop_param(port_dict, "security_groups")
    if new_security_groups is not None:
        if (Capabilities.TENANT_NETWORK_SG not in
                CONF.QUARK.environment_capabilities):
            if not STRATEGY.is_provider_network(port_db["network_id"]):
                raise q_exc.TenantNetworkSecurityGroupRulesNotEnabled()

    if new_security_groups is not None and not port_db["device_id"]:
        raise q_exc.SecurityGroupsRequireDevice()

    group_ids, security_group_mods = _make_security_group_list(
        context, new_security_groups)
    quota.QUOTAS.limit_check(context, context.tenant_id,
                             security_groups_per_port=len(group_ids))

    if fixed_ips is not None:
        # NOTE(mdietz): we want full control over IPAM since
        #              we're allocating by subnet instead of
        #              network.
        ipam_driver = ipam.IPAM_REGISTRY.get_strategy(
            ipam.QuarkIpamANY.get_name())

        addresses, subnet_ids = [], []
        ip_addresses = {}

        for fixed_ip in fixed_ips:
            subnet_id = fixed_ip.get("subnet_id")
            ip_address = fixed_ip.get("ip_address")
            if not (subnet_id or ip_address):
                raise n_exc.BadRequest(
                    resource="fixed_ips",
                    msg="subnet_id or ip_address required")

            if ip_address and not subnet_id:
                raise n_exc.BadRequest(
                    resource="fixed_ips",
                    msg="subnet_id required for ip_address allocation")

            if subnet_id and ip_address:
                ip_netaddr = None
                try:
                    ip_netaddr = netaddr.IPAddress(ip_address).ipv6()
                except netaddr.AddrFormatError:
                    raise n_exc.InvalidInput(
                        error_message="Invalid format provided for ip_address")
                ip_addresses[ip_netaddr] = subnet_id
            else:
                subnet_ids.append(subnet_id)

        port_ips = set([netaddr.IPAddress(int(a["address"]))
                        for a in port_db["ip_addresses"]])
        new_ips = set([a for a in ip_addresses.keys()])

        ips_to_allocate = list(new_ips - port_ips)
        ips_to_deallocate = list(port_ips - new_ips)

        for ip in ips_to_allocate:
            if ip in ip_addresses:
                # NOTE: Fix for RM10187 - we were losing the list of IPs if
                #       more than one IP was to be allocated. Track an
                #       aggregate list instead, and add it to the running total
                #       after each allocate
                allocated = []
                ipam_driver.allocate_ip_address(
                    context, allocated, port_db["network_id"],
                    port_db["id"], reuse_after=None, ip_addresses=[ip],
                    subnets=[ip_addresses[ip]])
                addresses.extend(allocated)

        for ip in ips_to_deallocate:
            ipam_driver.deallocate_ips_by_port(
                context, port_db, ip_address=ip)

        for subnet_id in subnet_ids:
            ipam_driver.allocate_ip_address(
                context, addresses, port_db["network_id"], port_db["id"],
                reuse_after=CONF.QUARK.ipam_reuse_after,
                subnets=[subnet_id])

        # Need to return all existing addresses and the new ones
        if addresses:
            port_dict["addresses"] = port_db["ip_addresses"]
            port_dict["addresses"].extend(addresses)

    # NOTE(morgabra) Updating network_plugin on port objects is explicitly
    # disallowed in the api, so we use whatever exists in the db.
    net_driver = _get_net_driver(port_db.network, port=port_db)
    base_net_driver = _get_net_driver(port_db.network)

    # TODO(anyone): What do we want to have happen here if this fails? Is it
    #               ok to continue to keep the IPs but fail to apply security
    #               groups? Is there a clean way to have a multi-status? Since
    #               we're in a beta-y status, I'm going to let this sit for
    #               a future patch where we have time to solve it well.
    kwargs = {}
    if new_security_groups is not None:
        kwargs["security_groups"] = security_group_mods
    net_driver.update_port(context, port_id=port_db["backend_key"],
                           mac_address=port_db["mac_address"],
                           device_id=port_db["device_id"],
                           base_net_driver=base_net_driver,
                           **kwargs)

    port_dict["security_groups"] = security_group_mods

    with context.session.begin():
        port = db_api.port_update(context, port_db, **port_dict)

    # NOTE(mdietz): fix for issue 112, we wanted the IPs to be in
    #              allocated_at order, so get a fresh object every time
    if port_db in context.session:
        context.session.expunge(port_db)
    port_db = db_api.port_find(context, id=id, scope=db_api.ONE)

    return v._make_port_dict(port_db)


def get_port(context, id, fields=None):
    """Retrieve a port.

    : param context: neutron api request context
    : param id: UUID representing the port to fetch.
    : param fields: a list of strings that are valid keys in a
        port dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_port %s for tenant %s fields %s" %
             (id, context.tenant_id, fields))
    results = db_api.port_find(context, id=id, fields=fields,
                               scope=db_api.ONE)

    if not results:
        raise n_exc.PortNotFound(port_id=id)

    return v._make_port_dict(results)


def get_ports(context, limit=None, sorts=None, marker=None, page_reverse=False,
              filters=None, fields=None):
    """Retrieve a list of ports.

    The contents of the list depends on the identity of the user
    making the request (as indicated by the context) as well as any
    filters.
    : param context: neutron api request context
    : param filters: a dictionary with keys that are valid keys for
        a port as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.
    : param fields: a list of strings that are valid keys in a
        port dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_ports for tenant %s filters %s fields %s" %
             (context.tenant_id, filters, fields))
    if filters is None:
        filters = {}

    if "ip_address" in filters:
        if not context.is_admin:
            raise n_exc.NotAuthorized()
        ips = []
        try:
            ips = [netaddr.IPAddress(ip) for ip in filters.pop("ip_address")]
        except netaddr.AddrFormatError:
            raise n_exc.InvalidInput(
                error_message="Invalid format provided for ip_address")
        query = db_api.port_find_by_ip_address(context, ip_address=ips,
                                               scope=db_api.ALL, **filters)
        ports = []
        for ip in query:
            ports.extend(ip.ports)
    else:
        ports = db_api.port_find(context, limit, sorts, marker,
                                 fields=fields, join_security_groups=True,
                                 **filters)
    return v._make_ports_list(ports, fields)


def get_ports_count(context, filters=None):
    """Return the number of ports.

    The result depends on the identity of the user making the request
    (as indicated by the context) as well as any filters.
    : param context: neutron api request context
    : param filters: a dictionary with keys that are valid keys for
        a port as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictionary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.

    NOTE: this method is optional, as it was not part of the originally
          defined plugin API.
    """
    LOG.info("get_ports_count for tenant %s filters %s" %
             (context.tenant_id, filters))
    return db_api.port_count_all(context, join_security_groups=True, **filters)


def delete_port(context, id):
    """Delete a port.

    : param context: neutron api request context
    : param id: UUID representing the port to delete.
    """
    LOG.info("delete_port %s for tenant %s" % (id, context.tenant_id))

    port = db_api.port_find(context, id=id, scope=db_api.ONE)
    if not port:
        raise n_exc.PortNotFound(port_id=id)

    if 'device_id' in port:  # false is weird, but ignore that
        LOG.info("delete_port %s for tenant %s has device %s" %
                 (id, context.tenant_id, port['device_id']))

    backend_key = port["backend_key"]
    mac_address = netaddr.EUI(port["mac_address"]).value
    ipam_driver = _get_ipam_driver(port["network"], port=port)
    ipam_driver.deallocate_mac_address(context, mac_address)
    ipam_driver.deallocate_ips_by_port(
        context, port, ipam_reuse_after=CONF.QUARK.ipam_reuse_after)

    net_driver = _get_net_driver(port["network"], port=port)
    base_net_driver = _get_net_driver(port["network"])
    net_driver.delete_port(context, backend_key, device_id=port["device_id"],
                           mac_address=port["mac_address"],
                           base_net_driver=base_net_driver)

    with context.session.begin():
        db_api.port_delete(context, port)


def _diag_port(context, port, fields):
    p = v._make_port_dict(port)
    net_driver = _get_net_driver(port.network, port=port)
    if 'config' in fields:
        p.update(net_driver.diag_port(
            context, port["backend_key"], get_status='status' in fields))
    return p


def diagnose_port(context, id, fields):
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    if id == "*":
        return {'ports': [_diag_port(context, port, fields) for
                port in db_api.port_find(context).all()]}
    db_port = db_api.port_find(context, id=id, scope=db_api.ONE)
    if not db_port:
        raise n_exc.PortNotFound(port_id=id)
    port = _diag_port(context, db_port, fields)
    return {'ports': port}


def _make_security_group_list(context, group_ids):
    if not group_ids or not utils.attr_specified(group_ids):
        return ([], [])
    group_ids = list(set(group_ids))
    groups = []
    for gid in group_ids:
        group = db_api.security_group_find(context, id=gid,
                                           scope=db_api.ONE)
        if not group:
            raise sg_ext.SecurityGroupNotFound(id=gid)
        groups.append(group)
    return (group_ids, groups)



# Copyright (c) 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from neutron._i18n import _
from oslo_config import cfg

CONF = cfg.CONF


quark_opts = [
    cfg.StrOpt('net_driver',
               default='quark.drivers.base.BaseDriver',
               help=_('The client to use to talk to the backend')),
    cfg.StrOpt('ipam_driver', default='quark.ipam.QuarkIpam',
               help=_('IPAM Implementation to use')),
    cfg.IntOpt('ipam_reuse_after', default=7200,
               help=_("Time in seconds til IP and MAC reuse"
                      "after deallocation.")),
    cfg.StrOpt("strategy_driver",
               default='quark.network_strategy.JSONStrategy',
               help=_("Tree of network assignment strategy")),
    cfg.StrOpt("default_network_type",
               default='BASE',
               help=_("Default network type to use when"
                      "none is provided")),
    cfg.StrOpt("default_ipam_strategy",
               default="ANY",
               help=_("Default IPAM strategy to use when"
                      "none is provided.")),
    cfg.BoolOpt("log_warnings", default=False,
                help=_("Redirect warnings to the logs."))
]


CONF.register_opts(quark_opts, "QUARK")

logging.captureWarnings(CONF.QUARK.log_warnings)



import datetime

import netaddr
from oslo_utils import timeutils

from quark.db import api as db_api
from quark.plugin_modules import ip_policies
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class IPReallocateMixin(object):
    REUSE_AFTER = 300

    def insert_network(self):
        tenant_id = "foobar"
        network = {"tenant_id": tenant_id}
        network_db = db_api.network_create(self.context, **network)
        self.context.session.flush()
        return network_db

    def insert_subnet(self, network_db, cidr, do_not_use=False):
        subnet = {"network": network_db,
                  "cidr": cidr,
                  "ip_version": netaddr.IPNetwork(cidr).version,
                  "do_not_use": do_not_use}
        subnet_db = db_api.subnet_create(self.context, **subnet)
        self.context.session.flush()
        return subnet_db

    def insert_ip_address(self, ip_address, network_db, subnet_db):
        ip_address_db = db_api.ip_address_create(
            self.context,
            address=ip_address,
            version=ip_address.version,
            subnet_id=subnet_db["id"] if subnet_db else None,
            network_id=network_db["id"])

        ip_address_db["_deallocated"] = True
        ip_address_db["deallocated_at"] = (
            timeutils.utcnow() - datetime.timedelta(seconds=self.REUSE_AFTER))
        self.context.session.add(ip_address_db)
        self.context.session.flush()

        return ip_address_db

    def insert_transaction(self):
        with self.context.session.begin():
            transaction = db_api.transaction_create(self.context)
        return transaction

    def default_case(self):
        self.network_db = self.insert_network()
        self.subnet_v4_db = self.insert_subnet(
            self.network_db, "192.168.0.0/24")
        self.subnet_v6_db = self.insert_subnet(
            self.network_db, "ffcc::/7")
        self.ip_address_v4 = netaddr.IPAddress("192.168.0.1")
        self.insert_ip_address(self.ip_address_v4, self.network_db,
                               self.subnet_v4_db)
        self.ip_address_v6 = netaddr.IPAddress("ffcc::1")
        self.insert_ip_address(self.ip_address_v6, self.network_db,
                               self.subnet_v6_db)
        self.transaction = self.insert_transaction()

    def insert_default_ip_policy(self, subnet_db):
        cidrs = []
        ip_policies.ensure_default_policy(cidrs, [subnet_db])
        subnet_db["ip_policy"] = db_api.ip_policy_create(
            self.context, exclude=cidrs)
        self.context.session.add(subnet_db)
        self.context.session.flush()


class QuarkIPReallocateFunctionalTest(MySqlBaseFunctionalTest,
                                      IPReallocateMixin):
    def setUp(self):
        super(QuarkIPReallocateFunctionalTest, self).setUp()
        self.default_case()

    def test_normal_v4(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

    def test_normal_v6(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 6,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

    def test_ip_address_specified_deallocated_None(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "ip_address": self.ip_address_v4,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

    def test_subnet_ids_specified(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
            "subnet_id": [self.subnet_v4_db["id"]]
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

    def test_reuse_after_not_time_yet(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER * 2,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertFalse(reallocated)

    def test_normal_one_of_multiple_potential_ip_addresses(self):
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)


class QuarkIPReallocateFindTest(MySqlBaseFunctionalTest, IPReallocateMixin):
    def test_normal_v4(self):
        self.default_case()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertEqual(updated_address["address"],
                         int(self.ip_address_v4.ipv6()))

    def test_normal_v6(self):
        self.default_case()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 6,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertEqual(updated_address["address"],
                         int(self.ip_address_v6.ipv6()))

    def test_address_not_found(self):
        self.transaction = self.insert_transaction()

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertIsNone(updated_address)

    def test_subnet_null(self):
        self.network_db = self.insert_network()
        self.subnet_v4_db = self.insert_subnet(
            self.network_db, "192.168.0.0/24")
        self.ip_address_v4 = netaddr.IPAddress("192.168.0.1")
        self.insert_ip_address(self.ip_address_v4, self.network_db, None)
        self.transaction = self.insert_transaction()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertIsNone(updated_address)

    def test_subnet_do_not_use(self):
        self.network_db = self.insert_network()
        self.subnet_v4_db = self.insert_subnet(
            self.network_db, "192.168.0.0/24", do_not_use=True)
        self.ip_address_v4 = netaddr.IPAddress("192.168.0.1")
        self.insert_ip_address(self.ip_address_v4, self.network_db,
                               self.subnet_v4_db)
        self.transaction = self.insert_transaction()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertIsNone(updated_address)

    def test_policy_violation(self):
        self.network_db = self.insert_network()
        self.subnet_v4_db = self.insert_subnet(
            self.network_db, "192.168.0.0/24")
        self.ip_address_v4 = netaddr.IPAddress("192.168.0.0")
        ip_address_db = self.insert_ip_address(self.ip_address_v4,
                                               self.network_db,
                                               self.subnet_v4_db)
        self.transaction = self.insert_transaction()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        self.insert_default_ip_policy(self.subnet_v4_db)
        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertIsNone(updated_address)

        self.context.session.flush()
        self.assertIsNone(db_api.ip_address_find(self.context,
                                                 id=ip_address_db.id,
                                                 scope=db_api.ONE))

    def test_address_not_in_cidr(self):
        self.network_db = self.insert_network()
        self.subnet_v4_db = self.insert_subnet(
            self.network_db, "192.168.0.0/24")
        self.ip_address_v4 = netaddr.IPAddress("192.168.1.1")
        ip_address_db = self.insert_ip_address(self.ip_address_v4,
                                               self.network_db,
                                               self.subnet_v4_db)
        self.transaction = self.insert_transaction()
        ip_kwargs = {
            "network_id": self.network_db["id"],
            "reuse_after": self.REUSE_AFTER,
            "deallocated": True,
            "version": 4,
        }
        reallocated = db_api.ip_address_reallocate(
            self.context,
            {"transaction_id": self.transaction.id},
            **ip_kwargs)
        self.assertTrue(reallocated)

        updated_address = db_api.ip_address_reallocate_find(
            self.context, self.transaction.id)
        self.assertIsNone(updated_address)

        self.context.session.flush()
        self.assertIsNone(db_api.ip_address_find(self.context,
                                                 id=ip_address_db.id,
                                                 scope=db_api.ONE))



"""indexaroo

Revision ID: 1bdc1b574beb
Revises: 79d90e62d88
Create Date: 2015-05-20 21:39:19.348638

"""

# revision identifiers, used by Alembic.
revision = '1bdc1b574beb'
down_revision = '79d90e62d88'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_quark_networks_tenant_id'),
                    'quark_networks',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_networks_tenant_id_name'),
                    'quark_networks',
                    ['tenant_id', 'name'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_tenant_id'),
                    'quark_subnets',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_network_id_tenant_id'),
                    'quark_subnets',
                    ['network_id', 'tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_subnets_network_id'),
                    'quark_subnets',
                    ['network_id'],
                    unique=False)
    op.create_index(op.f('ix_quotas_tenant_id'),
                    'quotas',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_tenant_id'),
                    'quark_ports',
                    ['tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_network_id_tenant_id'),
                    'quark_ports',
                    ['network_id', 'tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_name_tenant_id'),
                    'quark_ports',
                    ['name', 'tenant_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_mac_address_ranges_selection'),
                    'quark_mac_address_ranges',
                    ['next_auto_assign_mac', 'do_not_use'],
                    unique=False)
    op.create_index(op.f('ix_quark_mac_addresses_reallocation'),
                    'quark_mac_addresses',
                    ['deallocated_at', 'deallocated'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_quark_networks_tenant_id'),
                  table_name='quark_networks')
    op.drop_index(op.f('ix_quark_networks_tenant_id_name'),
                  table_name='quark_networks')

    op.drop_index(op.f('ix_quark_subnets_tenant_id'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_subnets_network_id_tenant_id'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_subnets_network_id'),
                  table_name='quark_subnets')

    op.drop_index(op.f('ix_quark_ports_tenant_id'),
                  table_name='ports')
    op.drop_index(op.f('ix_quark_ports_network_id_tenant_id'),
                  table_name='quark_ports')
    op.drop_index(op.f('ix_quark_ports_name_tenant_id'),
                  table_name='quark_ports')

    op.drop_index(op.f('ix_quotas_tenant_id'),
                  table_name='quotas')
    op.drop_index(op.f('ix_quark_mac_address_ranges_selection'),
                  table_name='quark_mac_address_ranges')
    op.drop_index(op.f('ix_quark_mac_addresses_reallocation'),
                  table_name='quark_mac_addresses')



"""Initial version

Revision ID: 1817eef6373c
Revises: None
Create Date: 2014-03-11 04:57:08.498604

"""

# revision identifiers, used by Alembic.
revision = '1817eef6373c'
down_revision = None

from alembic import op
import sqlalchemy as sa

from quark.db.custom_types import INET


def upgrade():
    op.create_table(
        'quark_mac_address_ranges',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('cidr', sa.String(length=255), nullable=False),
        sa.Column('first_address', sa.BigInteger(), nullable=False),
        sa.Column('last_address', sa.BigInteger(), nullable=False),
        sa.Column('next_auto_assign_mac', sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_tag_associations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('discriminator', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_ip_policy',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_security_groups',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('description', sa.String(length=255), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_security_groups_tenant_id'),
        'quark_security_groups',
        ['tenant_id'],
        unique=False)
    op.create_table(
        'quark_tags',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('association_uuid', sa.String(length=36), nullable=False),
        sa.Column('tag', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['association_uuid'],
                                [u'quark_tag_associations.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_security_group_rule',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('group_id', sa.String(length=36), nullable=False),
        sa.Column('direction', sa.String(length=10), nullable=False),
        sa.Column('ethertype', sa.String(length=4), nullable=False),
        sa.Column('port_range_max', sa.Integer(), nullable=True),
        sa.Column('port_range_min', sa.Integer(), nullable=True),
        sa.Column('protocol', sa.Integer(), nullable=True),
        sa.Column('remote_ip_prefix', sa.String(length=22), nullable=True),
        sa.Column('remote_group_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['group_id'], ['quark_security_groups.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_mac_addresses',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('address', sa.BigInteger(), nullable=False),
        sa.Column('mac_address_range_id', sa.String(length=36),
                  nullable=False),
        sa.Column('deallocated', sa.Boolean(), nullable=True),
        sa.Column('deallocated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['mac_address_range_id'],
                                ['quark_mac_address_ranges.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('address'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_mac_addresses_deallocated_at'),
        'quark_mac_addresses',
        ['deallocated_at'],
        unique=False)
    op.create_table(
        'quark_ip_policy_cidrs',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('ip_policy_id', sa.String(length=36), nullable=True),
        sa.Column('cidr', sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(['ip_policy_id'],
                                ['quark_ip_policy.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_networks',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('ip_policy_id', sa.String(length=36), nullable=True),
        sa.Column('network_plugin', sa.String(length=36), nullable=True),
        sa.Column('ipam_strategy', sa.String(length=255), nullable=True),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['ip_policy_id'], ['quark_ip_policy.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(op.f('ix_quark_networks_tenant_id'),
                    'quark_networks',
                    ['tenant_id'],
                    unique=False)
    op.create_table(
        'quark_ports',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('admin_state_up', sa.Boolean(), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('backend_key', sa.String(length=36), nullable=False),
        sa.Column('mac_address', sa.BigInteger(), nullable=True),
        sa.Column('device_id', sa.String(length=255), nullable=False),
        sa.Column('device_owner', sa.String(length=255), nullable=True),
        sa.Column('bridge', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['network_id'], ['quark_networks.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index('idx_ports_1',
                    'quark_ports',
                    ['device_id', 'tenant_id'],
                    unique=False)
    op.create_index('idx_ports_2',
                    'quark_ports',
                    ['device_owner', 'network_id'],
                    unique=False)
    op.create_index('idx_ports_3', 'quark_ports', ['tenant_id'], unique=False)
    op.create_index(op.f('ix_quark_ports_device_id'),
                    'quark_ports',
                    ['device_id'],
                    unique=False)
    op.create_index(op.f('ix_quark_ports_name'), 'quark_ports', ['name'],
                    unique=False)
    op.create_table(
        'quark_subnets',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('_cidr', sa.String(length=64), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('segment_id', sa.String(length=255), nullable=True),
        sa.Column('first_ip', INET(), nullable=True),
        sa.Column('last_ip', INET(), nullable=True),
        sa.Column('ip_version', sa.Integer(), nullable=True),
        sa.Column('next_auto_assign_ip', INET(), nullable=True),
        sa.Column('enable_dhcp', sa.Boolean(), nullable=True),
        sa.Column('ip_policy_id', sa.String(length=36), nullable=True),
        sa.Column('do_not_use', sa.Boolean(), nullable=True),
        sa.Column('tag_association_uuid', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['ip_policy_id'], ['quark_ip_policy.id'], ),
        sa.ForeignKeyConstraint(['network_id'], ['quark_networks.id'], ),
        sa.ForeignKeyConstraint(['tag_association_uuid'],
                                [u'quark_tag_associations.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_subnets_segment_id'),
        'quark_subnets',
        ['segment_id'],
        unique=False)
    op.create_index(
        op.f('ix_quark_subnets_tenant_id'),
        'quark_subnets',
        ['tenant_id'],
        unique=False)
    op.create_table(
        'quark_routes',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('cidr', sa.String(length=64), nullable=True),
        sa.Column('gateway', sa.String(length=64), nullable=True),
        sa.Column('subnet_id', sa.String(length=36), nullable=True),
        sa.Column('tag_association_uuid', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['subnet_id'], ['quark_subnets.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['tag_association_uuid'],
                                [u'quark_tag_associations.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_port_security_group_associations',
        sa.Column('port_id', sa.String(length=36), nullable=True),
        sa.Column('group_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['group_id'], ['quark_security_groups.id'], ),
        sa.ForeignKeyConstraint(['port_id'], ['quark_ports.id'], ),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_ip_addresses',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('address_readable', sa.String(length=128), nullable=False),
        sa.Column('address', INET(), nullable=False),
        sa.Column('subnet_id', sa.String(length=36), nullable=True),
        sa.Column('network_id', sa.String(length=36), nullable=True),
        sa.Column('version', sa.Integer(), nullable=True),
        sa.Column('allocated_at', sa.DateTime(), nullable=True),
        sa.Column('_deallocated', sa.Boolean(), nullable=True),
        sa.Column('used_by_tenant_id', sa.String(length=255), nullable=True),
        sa.Column('deallocated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['network_id'], ['quark_networks.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['subnet_id'], ['quark_subnets.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('subnet_id', 'address', name='subnet_id_address'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_ip_addresses_address'),
        'quark_ip_addresses',
        ['address'],
        unique=False)
    op.create_index(
        op.f('ix_quark_ip_addresses_deallocated_at'),
        'quark_ip_addresses',
        ['deallocated_at'],
        unique=False)
    op.create_index(
        op.f('ix_quark_ip_addresses_version'),
        'quark_ip_addresses',
        ['version'],
        unique=False)
    op.create_table(
        'quark_dns_nameservers',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('ip', INET(), nullable=True),
        sa.Column('subnet_id', sa.String(length=36), nullable=True),
        sa.Column('tag_association_uuid', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['subnet_id'], ['quark_subnets.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['tag_association_uuid'],
                                [u'quark_tag_associations.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_port_ip_address_associations',
        sa.Column('port_id', sa.String(length=36), nullable=True),
        sa.Column('ip_address_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['ip_address_id'], ['quark_ip_addresses.id'],),
        sa.ForeignKeyConstraint(['port_id'], ['quark_ports.id'], ),
        mysql_engine='InnoDB')
    op.create_table(
        'quark_nvp_driver_qos',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('display_name', sa.String(length=255), nullable=False),
        sa.Column('max_bandwidth_rate', sa.Integer(), nullable=False),
        sa.Column('min_bandwidth_rate', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_table(
        'quotas',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('resource', sa.String(length=255), nullable=True),
        sa.Column('limit', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quotas_tenant_id'),
        'quotas',
        ['tenant_id'],
        unique=False)
    op.create_table(
        'quark_nvp_driver_lswitch',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('nvp_id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('display_name', sa.String(length=255), nullable=True),
        sa.Column('port_count', sa.Integer(), nullable=True),
        sa.Column('transport_zone', sa.String(length=36), nullable=True),
        sa.Column('transport_connector', sa.String(length=20), nullable=True),
        sa.Column('segment_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_nvp_driver_lswitch_network_id'),
        'quark_nvp_driver_lswitch',
        ['network_id'],
        unique=False)
    op.create_index(
        op.f('ix_quark_nvp_driver_lswitch_nvp_id'),
        'quark_nvp_driver_lswitch',
        ['nvp_id'],
        unique=False)
    op.create_table(
        'quark_nvp_driver_security_profile',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('nvp_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_nvp_driver_security_profile_nvp_id'),
        'quark_nvp_driver_security_profile',
        ['nvp_id'],
        unique=False)
    op.create_table(
        'quark_nvp_driver_lswitchport',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('switch_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['switch_id'],
                                ['quark_nvp_driver_lswitch.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')
    op.create_index(
        op.f('ix_quark_nvp_driver_lswitchport_port_id'),
        'quark_nvp_driver_lswitchport',
        ['port_id'],
        unique=False)


def downgrade():
    op.drop_table('quark_port_ip_address_associations')
    op.drop_table('quark_dns_nameservers')
    op.drop_index(op.f('ix_quark_ip_addresses_version'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_ip_addresses_deallocated_at'),
                  table_name='quark_ip_addresses')
    op.drop_index(op.f('ix_quark_ip_addresses_address'),
                  table_name='quark_ip_addresses')
    op.drop_table('quark_ip_addresses')
    op.drop_table('quark_port_security_group_associations')
    op.drop_table('quark_routes')
    op.drop_index(op.f('ix_quark_subnets_tenant_id'),
                  table_name='quark_subnets')
    op.drop_index(op.f('ix_quark_subnets_segment_id'),
                  table_name='quark_subnets')
    op.drop_table('quark_subnets')
    op.drop_index(op.f('ix_quark_ports_name'), table_name='quark_ports')
    op.drop_index(op.f('ix_quark_ports_device_id'), table_name='quark_ports')
    op.drop_index('idx_ports_3', table_name='quark_ports')
    op.drop_index('idx_ports_2', table_name='quark_ports')
    op.drop_index('idx_ports_1', table_name='quark_ports')
    op.drop_table('quark_ports')
    op.drop_index(op.f('ix_quark_networks_tenant_id'),
                  table_name='quark_networks')
    op.drop_table('quark_networks')
    op.drop_table('quark_ip_policy_cidrs')
    op.drop_index(op.f('ix_quark_mac_addresses_deallocated_at'),
                  table_name='quark_mac_addresses')
    op.drop_table('quark_mac_addresses')
    op.drop_table('quark_security_group_rule')
    op.drop_table('quark_tags')
    op.drop_index(op.f('ix_quark_security_groups_tenant_id'),
                  table_name='quark_security_groups')
    op.drop_table('quark_security_groups')
    op.drop_table('quark_ip_policy')
    op.drop_table('quark_tag_associations')
    op.drop_table('quark_mac_address_ranges')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitchport_port_id'),
                  table_name='quark_nvp_driver_lswitchport')
    op.drop_table('quark_nvp_driver_lswitchport')
    op.drop_index(op.f('ix_quark_nvp_driver_security_profile_nvp_id'),
                  table_name='quark_nvp_driver_security_profile')
    op.drop_table('quark_nvp_driver_security_profile')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitch_nvp_id'),
                  table_name='quark_nvp_driver_lswitch')
    op.drop_index(op.f('ix_quark_nvp_driver_lswitch_network_id'),
                  table_name='quark_nvp_driver_lswitch')
    op.drop_table('quark_nvp_driver_lswitch')
    op.drop_index(op.f('ix_quotas_tenant_id'), table_name='quotas')
    op.drop_table('quotas')
    op.drop_table('quark_nvp_driver_qos')



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import defaultdict
import datetime
import json
import sys

import netaddr
from neutron.common import config
from neutron.db import api as neutron_db_api
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from sqlalchemy import and_, or_, func, not_

from quark.db import models

LOG = logging.getLogger(__name__)


def main():
    config.init(sys.argv[1:])
    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))
    config.setup_logging()

    public_network_id = "00000000-0000-0000-0000-000000000000"
    ip_availability = get_ip_availability(network_id=public_network_id,
                                          ip_version=4)
    print(json.dumps(ip_availability))


def get_ip_availability(**kwargs):
    LOG.debug("Begin querying %s" % kwargs)
    used_ips = get_used_ips(neutron_db_api.get_session(use_slave=True),
                            **kwargs)
    unused_ips = get_unused_ips(neutron_db_api.get_session(use_slave=True),
                                used_ips, **kwargs)
    LOG.debug("End querying")
    return dict(used=used_ips, unused=unused_ips)


def _convert_kwargs_values_into_tuples(f):
    def wrapped(*args, **kwargs):
        for key, value in kwargs.iteritems():
            if value is None:
                continue
            if not isinstance(value, list) and not isinstance(value, tuple):
                kwargs[key] = (value,)
        return f(*args, **kwargs)
    return wrapped


@_convert_kwargs_values_into_tuples
def _filter(query,
            network_id=None, segment_id=None, ip_version=None, subnet_id=None):
    query = query.filter(or_(models.Subnet.do_not_use.is_(None),
                             models.Subnet.do_not_use == 0))
    if network_id is not None:
        query = query.filter(models.Subnet.network_id.in_(network_id))
    if segment_id is not None:
        query = query.filter(models.Subnet.segment_id.in_(segment_id))
    if ip_version is not None:
        query = query.filter(models.Subnet.ip_version.in_(ip_version))
    if subnet_id is not None:
        query = query.filter(models.Subnet.id.in_(subnet_id))
    return query


def get_used_ips(session, **kwargs):
    """Returns dictionary with keys segment_id and value used IPs count.

    Used IP address count is determined by:
    - allocated IPs
    - deallocated IPs whose `deallocated_at` is within the `reuse_after`
    window compared to the present time, excluding IPs that are accounted for
    in the current IP policy (because IP policy is mutable and deallocated IPs
    are not checked nor deleted on IP policy creation, thus deallocated IPs
    that don't fit the current IP policy can exist in the neutron database).
    """
    LOG.debug("Getting used IPs...")
    with session.begin():
        query = session.query(
            models.Subnet.segment_id,
            func.count(models.IPAddress.address))
        query = query.group_by(models.Subnet.segment_id)
        query = _filter(query, **kwargs)

        reuse_window = timeutils.utcnow() - datetime.timedelta(
            seconds=cfg.CONF.QUARK.ipam_reuse_after)
        # NOTE(asadoughi): This is an outer join instead of a regular join
        # to include subnets with zero IP addresses in the database.
        query = query.outerjoin(
            models.IPAddress,
            and_(models.Subnet.id == models.IPAddress.subnet_id,
                 or_(not_(models.IPAddress.lock_id.is_(None)),
                     models.IPAddress._deallocated.is_(None),
                     models.IPAddress._deallocated == 0,
                     models.IPAddress.deallocated_at > reuse_window)))

        query = query.outerjoin(
            models.IPPolicyCIDR,
            and_(
                models.Subnet.ip_policy_id == models.IPPolicyCIDR.ip_policy_id,
                models.IPAddress.address >= models.IPPolicyCIDR.first_ip,
                models.IPAddress.address <= models.IPPolicyCIDR.last_ip))
        # NOTE(asadoughi): (address is allocated) OR
        # (address is deallocated and not inside subnet's IP policy)
        query = query.filter(or_(
            models.IPAddress._deallocated.is_(None),
            models.IPAddress._deallocated == 0,
            models.IPPolicyCIDR.id.is_(None)))

        ret = ((segment_id, address_count)
               for segment_id, address_count in query.all())
        return dict(ret)


def get_unused_ips(session, used_ips_counts, **kwargs):
    """Returns dictionary with key segment_id, and value unused IPs count.

    Unused IP address count is determined by:
    - adding subnet's cidr's size
    - subtracting IP policy exclusions on subnet
    - subtracting used ips per segment
    """
    LOG.debug("Getting unused IPs...")
    with session.begin():
        query = session.query(
            models.Subnet.segment_id,
            models.Subnet)
        query = _filter(query, **kwargs)
        query = query.group_by(models.Subnet.segment_id, models.Subnet.id)

        ret = defaultdict(int)
        for segment_id, subnet in query.all():
            net_size = netaddr.IPNetwork(subnet._cidr).size
            ip_policy = subnet["ip_policy"] or {"size": 0}
            ret[segment_id] += net_size - ip_policy["size"]

        for segment_id in used_ips_counts:
            ret[segment_id] -= used_ips_counts[segment_id]

        return ret



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import netaddr

from quark.db import api as db_api
import quark.ipam
import quark.plugin
from quark.tests.functional.base import BaseFunctionalTest


class QuarkIpamBaseFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super(QuarkIpamBaseFunctionalTest, self).setUp()
        self.plugin = quark.plugin.Plugin()


class QuarkTestIPFiltering(QuarkIpamBaseFunctionalTest):

    def setUp(self):
        super(QuarkTestIPFiltering, self).setUp()
        self.network = dict(name="public", tenant_id="fake")
        test_address1 = netaddr.IPAddress("192.168.1.100")
        test_address2 = netaddr.IPAddress("192.168.1.101")
        test_address3 = netaddr.IPAddress("192.168.1.102")
        ipnet = netaddr.IPNetwork("192.168.1.0/24")
        next_ip = ipnet.ipv6().first + 2
        self.subnet = dict(id=1, cidr="192.168.1.0/24",
                           next_auto_assign_ip=next_ip,
                           ip_policy=None, tenant_id="fake")

        self.ip_address1 = dict(address_readable="192.168.1.100",
                                subnet_id=1,
                                network_id=2, version=4,
                                address=test_address1, id=1)
        self.ip_address2 = dict(address_readable="192.168.1.101",
                                subnet_id=1,
                                network_id=2,
                                version=4,
                                address=test_address2,
                                id=2)
        self.ip_address3 = dict(address_readable="192.168.1.102",
                                subnet_id=1,
                                network_id=2,
                                version=4,
                                address=test_address3,
                                id=3)

    @contextlib.contextmanager
    def _stubs(self, network, subnet, ip_address1, ip_address2, ip_address3):
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            sub_mod = db_api.subnet_create(self.context, **subnet)
            # set tenant id to "123"
            ip_address1['network_id'] = net_mod.id
            ip_address1['subnet_id'] = sub_mod.id
            self.context.tenant_id = 123
            ip_address1 = db_api.ip_address_create(self.context, **ip_address1)
            # set tenant_id=456
            ip_address2['network_id'] = net_mod.id
            ip_address2['subnet_id'] = sub_mod.id
            self.context.tenant_id = 456
            ip_address2 = db_api.ip_address_create(self.context, **ip_address2)
            # set tenant id = "123" to test the list of IPs
            ip_address3['network_id'] = net_mod.id
            ip_address3['subnet_id'] = sub_mod.id
            self.context.tenant_id = 123
            ip_address3 = db_api.ip_address_create(self.context, **ip_address3)
        yield net_mod
        with self.context.session.begin():
            db_api.subnet_delete(self.context, sub_mod)
            db_api.network_delete(self.context, net_mod)

    def test_basic_ip_filtering_with_tenant_id(self):
        with self._stubs(self.network, self.subnet,
                         self.ip_address1,
                         self.ip_address2, self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context, tenant_id="123")
            self.assertEqual(2, len(res))
            id = int(res[0].get("id"))
            self.assertEqual(self.ip_address1["id"], id)
            res = self.plugin.get_ip_addresses(self.context, tenant_id="456")
            self.assertEqual(1, len(res))
            id = int(res[0].get("id"))
            self.assertEqual(self.ip_address2["id"], id)

    def test_basic_ip_filtering_with_same_tenant_id_with_different_ip(self):
        with self._stubs(self.network, self.subnet,
                         self.ip_address1, self.ip_address2,
                         self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context, tenant_id="123")
            self.assertEqual(2, len(res))
            id1 = int(res[0].get("id"))
            id3 = int(res[1].get("id"))
            self.assertEqual(self.ip_address1["id"], id1)
            self.assertEqual(self.ip_address3["id"], id3)

    def test_basic_ip_filtering_without_tenant_id(self):
        with self._stubs(self.network, self.subnet, self.ip_address1,
                         self.ip_address2, self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context)
            self.assertEqual(2, len(res))

    def test_basic_ip_filtering_with_tenant_id_without_ip(self):
        with self._stubs(self.network, self.subnet,
                         self.ip_address1,
                         self.ip_address2,
                         self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context, tenant_id="1234")
            self.assertEqual(0, len(res))

    def test_basic_ip_filtering_with_used_by_tenant_id(self):
        with self._stubs(self.network, self.subnet, self.ip_address1,
                         self.ip_address2, self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context,
                                               used_by_tenant_id="123")
            self.assertEqual(2, len(res))
            id = int(res[0].get("id"))
            self.assertEqual(self.ip_address1["id"], id)
            res = self.plugin.get_ip_addresses(self.context,
                                               used_by_tenant_id="456")
            self.assertEqual(0, len(res))

    def test_filtering_with_same_used_by_tenant_id_with_different_ip(self):
        with self._stubs(self.network, self.subnet,
                         self.ip_address1, self.ip_address2,
                         self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context,
                                               used_by_tenant_id="123")
            self.assertEqual(2, len(res))
            id1 = int(res[0].get("id"))
            id3 = int(res[1].get("id"))
            self.assertEqual(self.ip_address1["id"], id1)
            self.assertEqual(self.ip_address3["id"], id3)

    def test_basic_ip_filtering_with_used_by_tenant_id_without_ip(self):
        with self._stubs(self.network, self.subnet,
                         self.ip_address1,
                         self.ip_address2,
                         self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context,
                                               used_by_tenant_id="1234")
            self.assertEqual(0, len(res))

    def test_basic_ip_filtering_without_tenant_id_as_admin(self):
        with self._stubs(self.network, self.subnet, self.ip_address1,
                         self.ip_address2, self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context.elevated())
            self.assertEqual(3, len(res))

    def test_basic_ip_filtering_with_used_by_tenant_id_as_admin(self):
        with self._stubs(self.network, self.subnet, self.ip_address1,
                         self.ip_address2, self.ip_address3):
            res = self.plugin.get_ip_addresses(self.context.elevated(),
                                               used_by_tenant_id="123")
            self.assertEqual(2, len(res))
            id = int(res[0].get("id"))
            self.assertEqual(self.ip_address1["id"], id)
            res = self.plugin.get_ip_addresses(self.context.elevated(),
                                               used_by_tenant_id="456")
            self.assertEqual(1, len(res))



import json
import mock
import netaddr
from neutron import context
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import api as db_api
from quark import exceptions as qexceptions
import quark.ipam
from quark import network_strategy
import quark.plugin
import quark.plugin_modules.mac_address_ranges as macrng_api
from quark.tests.functional import base


class BaseFloatingIPTest(base.BaseFunctionalTest):

    FAKE_UNICORN_URL = 'http://unicorn.xxx'

    def _setup_mock_requests(self):
        cfg.CONF.set_override('floating_ip_base_url', self.FAKE_UNICORN_URL,
                              group='QUARK')
        patcher = mock.patch('quark.drivers.unicorn_driver.requests')
        self.mock_requests = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_requests.post.return_value.status_code = 200
        self.mock_requests.delete.return_value.status_code = 204
        self.mock_requests.put.return_value.status_code = 200

    def _build_expected_unicorn_request_body(self, floating_ip_address, ports,
                                             actual_body=None):
        if actual_body:
            # Since the port order is non-deterministic, we need to ensure
            # that the order is correct
            actual_port_ids = [endpoint['port']['uuid'] for endpoint in
                               actual_body['floating_ip']['endpoints']]
            reordered_ports = []
            for port_id in actual_port_ids:
                for port in ports:
                    if port['id'] == port_id:
                        reordered_ports.append(port)
            ports = reordered_ports
        endpoints = []
        for port in ports:
            fixed_ips = []
            for fixed_ip in port['fixed_ips']:
                fixed_ips.append({
                    'ip_address': fixed_ip['ip_address'],
                    'version': self.user_subnet['ip_version'],
                    'subnet_id': self.user_subnet['id'],
                    'cidr': self.user_subnet['cidr'],
                    'address_type': 'fixed'
                })
            port_mac = int(netaddr.EUI(port['mac_address'].replace(':', '-')))
            endpoints.append({
                'port': {
                    'uuid': port['id'],
                    'name': port['name'],
                    'network_uuid': port['network_id'],
                    'mac_address': port_mac,
                    'device_id': port['device_id'],
                    'device_owner': port['device_owner'],
                    'fixed_ip': fixed_ips
                },
                'private_ip': port['fixed_ips'][0]['ip_address']
            })
        body = {'public_ip': floating_ip_address,
                'endpoints': endpoints}
        return {'floating_ip': body}

    def setUp(self):
        super(BaseFloatingIPTest, self).setUp()
        self.public_net_id = "00000000-0000-0000-0000-000000000000"
        net_stat = '{"%s": {}}' % self.public_net_id
        cfg.CONF.set_override('default_net_strategy', net_stat, group='QUARK')
        old_strat = db_api.STRATEGY

        def reset_strategy():
            db_api.STRATEGY = old_strat

        db_api.STRATEGY = network_strategy.JSONStrategy()
        self.addCleanup(reset_strategy)
        admin_ctx = context.get_admin_context()
        self._setup_mock_requests()
        self.plugin = quark.plugin.Plugin()
        mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
        macrng_api.create_mac_address_range(admin_ctx, mac)
        with admin_ctx.session.begin():
            tenant = 'rackspace'
            floating_net = dict(name='publicnet', tenant_id=tenant,
                                id=self.public_net_id)
            self.floating_network = db_api.network_create(
                self.context, **floating_net)
            self.pub_net_cidr = "10.1.1.0/24"
            floating_subnet = dict(id=self.public_net_id,
                                   cidr=self.pub_net_cidr,
                                   ip_policy=None, tenant_id=tenant,
                                   segment_id='floating_ip',
                                   network_id=self.floating_network.id)
            self.floating_subnet = db_api.subnet_create(
                self.context, **floating_subnet)
        user_net = dict(name='user_network', tenant_id='fake')
        self.user_network = self.plugin.create_network(
            self.context, {'network': user_net})
        user_subnet = dict(cidr="192.168.1.0/24",
                           ip_policy=None, tenant_id="fake",
                           network_id=self.user_network['id'])
        self.user_subnet = self.plugin.create_subnet(
            self.context, {'subnet': user_subnet})
        user_port1 = dict(name='user_port1',
                          network_id=self.user_network['id'])
        self.user_port1 = self.plugin.create_port(
            self.context, {'port': user_port1})
        user_port2 = dict(name='user_port2',
                          network_id=self.user_network['id'])
        self.user_port2 = self.plugin.create_port(
            self.context, {'port': user_port2})


class TestFloatingIPs(BaseFloatingIPTest):

    def test_create(self):
        flip_req = dict(
            floating_network_id=self.floating_network['id'],
            port_id=self.user_port1['id']
        )
        flip_req = {'floatingip': flip_req}
        flip = self.plugin.create_floatingip(self.context, flip_req)
        self.assertIn(netaddr.IPAddress(flip['floating_ip_address']),
                      list(netaddr.IPNetwork(self.pub_net_cidr)))
        self.assertEqual(self.floating_network['id'],
                         flip['floating_network_id'])
        self.assertEqual(self.user_port1['id'], flip['port_id'])
        self.assertEqual(self.user_port1['fixed_ips'][0]['ip_address'],
                         flip['fixed_ip_address'])
        self.mock_requests.post.assert_called_once_with(
            self.FAKE_UNICORN_URL, data=mock.ANY, timeout=2
        )
        actual_body = json.loads(self.mock_requests.post.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            flip['floating_ip_address'], [self.user_port1]
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")
        get_flip = self.plugin.get_floatingip(self.context, flip['id'])
        self.assertEqual(flip['floating_ip_address'],
                         get_flip['floating_ip_address'])

    def test_create_raises_missing_key(self):
        flip_req = dict(
            floating_network_id=self.floating_network['id'],
            port_id=self.user_port1['id']
        )
        flip_req = {'floatingipz': flip_req}
        with self.assertRaises(n_exc.BadRequest):
            self.plugin.create_floatingip(self.context, flip_req)

    def test_update_floating_ip(self):
        floating_ip = dict(
            floating_network_id=self.floating_network.id,
            port_id=self.user_port1['id']
        )
        floating_ip = {'floatingip': floating_ip}
        flip = self.plugin.create_floatingip(self.context, floating_ip)
        fixed_ip_address2 = self.user_port2['fixed_ips'][0]['ip_address']
        floating_ip = dict(port_id=self.user_port2['id'],
                           fixed_ip_address=fixed_ip_address2)
        updated_flip = self.plugin.update_floatingip(
            self.context, flip['id'], {"floatingip": floating_ip})
        self.assertEqual(self.floating_network['id'],
                         updated_flip['floating_network_id'])
        self.assertEqual(updated_flip['floating_ip_address'],
                         flip['floating_ip_address'])
        self.assertEqual(self.user_port2['id'], updated_flip['port_id'])
        self.assertEqual(self.user_port2['fixed_ips'][0]['ip_address'],
                         updated_flip['fixed_ip_address'])
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 flip['floating_ip_address']])
        self.mock_requests.put.assert_called_once_with(
            expected_url, data=mock.ANY, timeout=2
        )
        actual_body = json.loads(self.mock_requests.put.call_args[1]['data'])
        unicorn_body = self._build_expected_unicorn_request_body(
            flip['floating_ip_address'], [self.user_port2]
        )
        self.assertEqual(unicorn_body, actual_body,
                         msg="Request to the unicorn API is not what is "
                             "expected.")
        get_flip = self.plugin.get_floatingip(self.context, flip['id'])
        self.assertEqual(flip['floating_ip_address'],
                         get_flip['floating_ip_address'])

    @mock.patch('quark.billing.notify')
    @mock.patch('quark.billing.build_payload', return_value={})
    def test_delete_floating_ip(self, notify, build_payload):
        floating_ip = dict(
            floating_network_id=self.floating_network.id,
            port_id=self.user_port1['id'],
            address_type='floating'
        )
        flip = self.plugin.create_floatingip(
            self.context, {"floatingip": floating_ip})
        self.plugin.delete_floatingip(self.context, flip['id'])
        expected_url = '/'.join([self.FAKE_UNICORN_URL,
                                 flip['floating_ip_address']])
        self.mock_requests.delete.assert_called_once_with(
            expected_url, timeout=2
        )
        self.assertRaises(qexceptions.FloatingIpNotFound,
                          self.plugin.get_floatingip, self.context, flip['id'])
        flips = self.plugin.get_floatingips(self.context)
        self.assertEqual(0, len(flips))



# Copyright 2014 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#

import json

import netaddr
from oslo_config import cfg
from oslo_log import log as logging

from quark.cache import redis_base
from quark.environment import Capabilities
from quark import exceptions as q_exc
from quark import protocols
from quark import utils


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
SECURITY_GROUP_RULE_KEY = "rules"
SECURITY_GROUP_HASH_ATTR = "security group rules"
SECURITY_GROUP_ACK = "security group ack"

ALL_V4 = netaddr.IPNetwork("::ffff:0.0.0.0/96")
ALL_V6 = netaddr.IPNetwork("::/0")


class SecurityGroupsClient(redis_base.ClientBase):
    def _convert_remote_network(self, remote_ip_prefix):
        # NOTE(mdietz): RM11364 - While a /0 is valid and should be supported,
        #               it breaks OVS to apply a /0 as the source or
        #               destination network.
        net = netaddr.IPNetwork(remote_ip_prefix).ipv6()
        if net.cidr == ALL_V4 or net.cidr == ALL_V6:
            return ''
        return str(net)

    def serialize_rules(self, rules):
        """Creates a payload for the redis server."""
        # TODO(mdietz): If/when we support other rule types, this comment
        #               will have to be revised.
        # Action and direction are static, for now. The implementation may
        # support 'deny' and 'egress' respectively in the future. We allow
        # the direction to be set to something else, technically, but current
        # plugin level call actually raises. It's supported here for unit
        # test purposes at this time
        serialized = []
        for rule in rules:
            direction = rule["direction"]
            source = ''
            destination = ''
            if rule.get("remote_ip_prefix"):
                prefix = rule["remote_ip_prefix"]
                if direction == "ingress":
                    source = self._convert_remote_network(prefix)
                else:
                    if (Capabilities.EGRESS not in
                            CONF.QUARK.environment_capabilities):
                        raise q_exc.EgressSecurityGroupRulesNotEnabled()
                    else:
                        destination = self._convert_remote_network(prefix)

            optional_fields = {}

            # NOTE(mdietz): this will expand as we add more protocols
            protocol_map = protocols.PROTOCOL_MAP[rule["ethertype"]]
            if rule["protocol"] == protocol_map["icmp"]:
                optional_fields["icmp type"] = rule["port_range_min"]
                optional_fields["icmp code"] = rule["port_range_max"]
            else:
                optional_fields["port start"] = rule["port_range_min"]
                optional_fields["port end"] = rule["port_range_max"]

            payload = {"ethertype": rule["ethertype"],
                       "protocol": rule["protocol"],
                       "source network": source,
                       "destination network": destination,
                       "action": "allow",
                       "direction": direction}
            payload.update(optional_fields)
            serialized.append(payload)
        return serialized

    def serialize_groups(self, groups):
        """Creates a payload for the redis server

        The rule schema is the following:

        REDIS KEY - port_device_id.port_mac_address/sg
        REDIS VALUE - A JSON dump of the following:

        port_mac_address must be lower-cased and stripped of non-alphanumeric
        characters

        {"id": "<arbitrary uuid>",
          "rules": [
            {"ethertype": <hexademical integer>,
             "protocol": <integer>,
             "port start": <integer>,  # optional
             "port end": <integer>,    # optional
             "icmp type": <integer>,   # optional
             "icmp code": <integer>,   # optional
             "source network": <string>,
             "destination network": <string>,
             "action": <string>,
             "direction": <string>},
          ],
          "security groups ack": <boolean>
        }

        Example:
        {"id": "004c6369-9f3d-4d33-b8f5-9416bf3567dd",
         "rules": [
           {"ethertype": 0x800,
            "protocol": "tcp",
            "port start": 1000,
            "port end": 1999,
            "source network": "10.10.10.0/24",
            "destination network": "",
            "action": "allow",
            "direction": "ingress"},
          ],
          "security groups ack": "true"
        }

        port start/end and icmp type/code are mutually exclusive pairs.
        """
        rules = []
        for group in groups:
            rules.extend(self.serialize_rules(group.rules))
        return rules

    def get_rules_for_port(self, device_id, mac_address):
        rules = self.get_field(
            self.vif_key(device_id, mac_address), SECURITY_GROUP_HASH_ATTR)
        if rules:
            return json.loads(rules)

    def apply_rules(self, device_id, mac_address, rules):
        """Writes a series of security group rules to a redis server."""
        LOG.info("Applying security group rules for device %s with MAC %s" %
                 (device_id, mac_address))

        rule_dict = {SECURITY_GROUP_RULE_KEY: rules}
        redis_key = self.vif_key(device_id, mac_address)
        # TODO(mdietz): Pipeline these. Requires some rewriting
        self.set_field(redis_key, SECURITY_GROUP_HASH_ATTR, rule_dict)
        self.set_field_raw(redis_key, SECURITY_GROUP_ACK, False)

    def delete_vif_rules(self, device_id, mac_address):
        # Redis HDEL command will ignore key safely if it doesn't exist
        self.delete_field(self.vif_key(device_id, mac_address),
                          SECURITY_GROUP_HASH_ATTR)
        self.delete_field(self.vif_key(device_id, mac_address),
                          SECURITY_GROUP_ACK)

    def delete_vif(self, device_id, mac_address):
        # Redis DEL command will ignore key safely if it doesn't exist
        self.delete_key(self.vif_key(device_id, mac_address))

    @utils.retry_loop(3)
    def get_security_group_states(self, interfaces):
        """Gets security groups for interfaces from Redis

        Returns a dictionary of xapi.VIFs with values of the current
        acknowledged status in Redis.

        States not explicitly handled:
        * ack key, no rules - This is the same as just tagging the VIF,
          the instance will be inaccessible
        * rules key, no ack - Nothing will happen, the VIF will
          not be tagged.
        """
        LOG.debug("Getting security groups from Redis for {0}".format(
            interfaces))
        interfaces = tuple(interfaces)
        vif_keys = [self.vif_key(vif.device_id, vif.mac_address)
                    for vif in interfaces]

        security_groups = self.get_fields(vif_keys, SECURITY_GROUP_ACK)

        ret = {}
        for vif, security_group_ack in zip(interfaces, security_groups):
            if security_group_ack:
                security_group_ack = security_group_ack.lower()
                if "true" in security_group_ack:
                    ret[vif] = True
                elif "false" in security_group_ack:
                    ret[vif] = False
                else:
                    LOG.debug("Skipping bad ack value %s" % security_group_ack)
        return ret

    @utils.retry_loop(3)
    def update_group_states_for_vifs(self, vifs, ack):
        """Updates security groups by setting the ack field"""
        vif_keys = [self.vif_key(vif.device_id, vif.mac_address)
                    for vif in vifs]
        self.set_fields(vif_keys, SECURITY_GROUP_ACK, ack)
